import json
import logging
import os
import sys
from functools import wraps
from logging.config import dictConfig
from logging.handlers import RotatingFileHandler
from threading import RLock

from fedot.core.utils import default_fedot_data_dir


class SingletonMeta(type):
    """
    This meta class can provide other classes with the Singleton pattern.
    It guarantees to create one and only class instance.
    Pass it to the metaclass parameter when defining your class as follows:

    class YourClassName(metaclass=SingletonMeta)
    """
    _instances = {}

    _lock: RLock = RLock()

    def __call__(cls, *args, **kwargs):
        with cls._lock:
            if cls not in cls._instances:
                instance = super().__call__(*args, **kwargs)
                cls._instances[cls] = instance
        return cls._instances[cls]


class LogManager(metaclass=SingletonMeta):
    __logger_dict = {}

    def __init__(self):
        pass

    def get_logger(self, name, config_file: str, log_file: str = None):
        if name not in self.__logger_dict.keys():
            self.__logger_dict[name] = logging.getLogger(name)
            if config_file != 'default':
                self._setup_logger_from_json_file(config_file)
            else:
                self._setup_default_logger(log_file, name)

        return self.__logger_dict[name]

    def _setup_default_logger(self, log_file, logger_name):
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        file_handler = RotatingFileHandler(log_file)
        file_handler.setLevel(logging.INFO)
        file_handler.setFormatter(formatter)
        console_handler = logging.StreamHandler(sys.stdout)
        self.__logger_dict[logger_name].setLevel(logging.INFO)
        self.__logger_dict[logger_name].addHandler(file_handler)
        self.__logger_dict[logger_name].addHandler(console_handler)

    def _setup_logger_from_json_file(self, config_file):
        """Setup logging configuration from file"""
        try:
            with open(config_file, 'rt') as file:
                config = json.load(file)
            dictConfig(config)
        except Exception as ex:
            raise Exception(f'Can not open the log config file because of {ex}')

    @property
    def debug(self):
        """Returns the information about available loggers"""
        debug_info = {
            'loggers_number': len(self.__logger_dict),
            'loggers_names': [self.__logger_dict.keys()],
            'loggers': [self.__logger_dict.values()]
        }
        return debug_info

    def clear_cache(self):
        self.__logger_dict.clear()


def default_log(logger_name: str,
                log_file=None) -> 'Log':
    """
    :param logger_name: string name for logger
    :param log_file: path to the file where log messages will be recorded to
    :return Log: Log object
    """
    if not log_file:
        log_file = os.path.join(default_fedot_data_dir(), 'log.log')
    log = Log(logger_name=logger_name,
              config_json_file='default',
              log_file=log_file)
    return log


class Log:
    """
    This class provides with basic logging object

    :param str logger_name: name of the logger object
    :param str config_json_file: json file with configuration for logger setup
    :param str log_file: file where log messages are recorded to
    """

    def __init__(self, logger_name: str,
                 config_json_file: str,
                 log_file: str = None):
        if not log_file:
            self.log_file = os.path.join(default_fedot_data_dir(), 'log.log')
        else:
            self.log_file = log_file

        self.name = logger_name
        self.config_file = config_json_file
        self.logger = LogManager().get_logger(logger_name,
                                              config_file=self.config_file,
                                              log_file=self.log_file)

    def info(self, message):
        """Record the INFO log massage"""
        self.logger.info(message)

    def debug(self, message):
        """Record the DEBUG log massage"""
        self.logger.debug(message)

    def warn(self, message):
        """Record the WARN log massage"""
        self.logger.warning(message)

    def error(self, message):
        """Record the ERROR log massage"""
        self.logger.error(message, exc_info=True)

    @property
    def handlers(self):
        return self.logger.handlers

    def release_handlers(self):
        """This function closes handlers of logger"""
        for handler in self.handlers:
            handler.close()

    def __getstate__(self):
        """
        Define the attributes to be pickled via deepcopy or pickle

        :return: dict: state
        """
        state = dict(self.__dict__)
        del state['logger']
        return state

    def __setstate__(self, state):
        """
        Restore an unpickled dict state and assign state items
        to the new instance’s dictionary.

        :param state: pickled class attributes
        """
        self.__dict__.update(state)
        self.logger = logging.getLogger(self.name)

    def __str__(self):
        return f'Log object for {self.name} module'

    def __repr__(self):
        return self.__str__()


def start_end_log_decorator(start_msg='Starting...', end_msg='Finished'):
    def decorator(method):
        @wraps(method)
        def wrapper(*args, **kwargs):
            args[0].log.info(f'{start_msg}')
            value = method(*args, **kwargs)
            args[0].log.info(f'{end_msg}')
            return value

        return wrapper

    return decorator



import datetime
from abc import ABC, abstractmethod
from typing import (Any, Callable, List, Optional)

from dataclasses import dataclass

from fedot.core.composer.chain import Chain
from fedot.core.composer.node import PrimaryNode, SecondaryNode
from fedot.core.log import Log, default_log
from fedot.core.models.data import InputData
from fedot.core.utils import ComparableEnum as Enum


@dataclass
class ComposerRequirements:
    """
    This dataclass is for defining the requirements of composition process

    :param primary: List of model types (str) for Primary Nodes
    :param secondary: List of model types (str) for Secondary Nodes
    :param max_lead_time: max time in minutes available for composition process
    :param max_depth: max depth of the result chain
    :param max_arity: maximal number of parent for node
    :param min_arity: minimal number of parent for node
    :param add_single_model_chains: allow to have chain with only one node
    """
    primary: List[str]
    secondary: List[str]
    max_lead_time: Optional[datetime.timedelta] = datetime.timedelta(minutes=5)
    max_depth: int = 3
    max_arity: int = 2
    min_arity: int = 2
    add_single_model_chains: bool = True

    def __post_init__(self):
        if self.max_depth < 0:
            raise ValueError(f'invalid max_depth value')
        if self.max_arity < 0:
            raise ValueError(f'invalid max_arity value')
        if self.min_arity < 0:
            raise ValueError(f'invalid min_arity value')


class Composer(ABC):
    """
    Base class used for receiving composite models via optimization
    :param metrics: metrics used to define the quality of found solution
    :param composer_requirements: requirements for composition process
    :param optimiser_parameters: parameters used by optimization process (i.e. GPComposerRequirements)
    :param initial_chain: defines the initial state of the population. If None then initial population is random.
    :param log: optional parameter for log oject
    """

    def __init__(self, metrics: Optional[Callable], composer_requirements: ComposerRequirements,
                 optimiser_parameters: Any = None, initial_chain: Optional[Chain] = None,
                 log: Log = None):
        self.history = None
        self.metrics = metrics
        self.composer_requirements = composer_requirements
        self.optimiser_parameters = optimiser_parameters
        self.initial_chain = initial_chain

        if not log:
            self.log = default_log(__name__)
        else:
            self.log = log

    @abstractmethod
    def compose_chain(self, data: InputData,
                      is_visualise: bool = False) -> Chain:
        """
        Base method to run the composition process

        :param data: data used for problem solving
        :param is_visualise: flag to enable visualization. Default False.
        :return: Chain object
        """
        raise NotImplementedError()


class DummyChainTypeEnum(Enum):
    flat = 1,
    hierarchical = 2


class DummyComposer(Composer):
    def __init__(self, dummy_chain_type: DummyChainTypeEnum,
                 composer_requirements: ComposerRequirements,
                 metrics: Optional[Callable],
                 initial_chain: Optional[Chain] = None):
        super().__init__(composer_requirements=composer_requirements, metrics=metrics,
                         initial_chain=initial_chain)
        self.dummy_chain_type = dummy_chain_type

    def compose_chain(self, data: InputData, is_visualise: bool = False) -> Chain:
        new_chain = Chain()

        if self.dummy_chain_type == DummyChainTypeEnum.hierarchical:
            # (y1, y2) -> y
            last_node = SecondaryNode(self.composer_requirements.secondary[0])

            for requirement_model in self.composer_requirements.primary:
                new_node = PrimaryNode(requirement_model)
                new_chain.add_node(new_node)
                last_node.nodes_from.append(new_node)
            new_chain.add_node(last_node)
        elif self.dummy_chain_type == DummyChainTypeEnum.flat:
            # (y1) -> (y2) -> y
            first_node = PrimaryNode(self.composer_requirements.primary[0])
            new_chain.add_node(first_node)
            prev_node = first_node
            for requirement_model in self.composer_requirements.secondary:
                new_node = SecondaryNode(requirement_model)
                new_node.nodes_from = [prev_node]
                prev_node = new_node
                new_chain.add_node(new_node)
        else:
            raise NotImplementedError()
        return new_chain



import datetime
from abc import abstractmethod, ABC


class Timer(ABC):
    def __init__(self, verbose=True):
        self.verbose = verbose
        self.process_terminated = False

    def __enter__(self):
        self.start = datetime.datetime.now()
        return self

    @abstractmethod
    def is_time_limit_reached(self, *args):
        raise NotImplementedError()

    def __exit__(self, *args):
        raise NotImplementedError()


class CompositionTimer(Timer):
    def __init__(self, verbose=True):
        super().__init__(verbose=verbose)

    @property
    def start_time(self):
        return self.start

    @property
    def spent_time(self) -> datetime.timedelta:
        return datetime.datetime.now() - self.start

    @property
    def minutes_from_start(self) -> float:
        return self.spent_time.seconds / 60.

    def _is_next_iteration_possible(self, time_constraint: float, generation_num: int = None) -> bool:
        minutes = self.minutes_from_start
        if generation_num is not None:
            possible = time_constraint > (minutes + (minutes / (generation_num + 1)))
        else:
            possible = time_constraint > minutes
        if not possible:
            self.process_terminated = True
        return possible

    def is_time_limit_reached(self, max_lead_time: datetime.timedelta, generation_num: int = None) -> bool:
        max_lead_time = 0 if max_lead_time.seconds < 0 else max_lead_time.seconds / 60.
        if max_lead_time:
            reached = not self._is_next_iteration_possible(generation_num=generation_num, time_constraint=max_lead_time)
        else:
            self.process_terminated = True
            reached = True
        return reached

    def __exit__(self, *args):
        if self.verbose:
            print(f'Composition time: {round(self.minutes_from_start, 3)} min')
            if self.process_terminated:
                print('Algorithm was terminated due to processing time limit')


class TunerTimer(Timer):
    def __init__(self, verbose=False):
        super().__init__(verbose=verbose)

    def is_time_limit_reached(self, limit) -> bool:
        if datetime.datetime.now() - self.start >= limit:
            self.process_terminated = True
            if self.verbose:
                print('Tuning completed because of the time limit reached')
        return self.process_terminated

    def __exit__(self, *args):
        return self.process_terminated



from copy import copy
from datetime import timedelta
from typing import Optional

import numpy as np
import tensorflow as tf

from fedot.core.models.data import InputData, OutputData
from fedot.core.models.evaluation.evaluation import EvaluationStrategy
from fedot.core.repository.tasks import extract_task_param

forecast_length = 1


# TODO inherit this and similar from custom strategy
class KerasForecastingStrategy(EvaluationStrategy):
    def __init__(self, model_type: str, params: Optional[dict] = None):
        self._init_lstm_model_functions(model_type)

        self.epochs = 10
        if params:
            self.epochs = params.get('epochs', self.epochs)

        super().__init__(model_type, params)

    def _init_lstm_model_functions(self, model_type):
        if model_type != 'lstm':
            raise ValueError(f'Impossible to obtain forecasting strategy for {model_type}')

    def fit(self, train_data: InputData):
        model = fit_lstm(train_data, epochs=self.epochs)
        return model

    def predict(self, trained_model, predict_data: InputData):
        return predict_lstm(trained_model, predict_data)

    def fit_tuned(self, train_data: InputData, iterations: int = 30,
                  max_lead_time: timedelta = timedelta(minutes=5)):
        raise NotImplementedError()


def _rmse_only_last(y_true, y_pred):
    """
    Computes rmse only on the last `prediction_len` values - forecasting
    """
    global forecast_length
    y_true = y_true[:, -forecast_length:]
    y_pred = y_pred[:, -forecast_length:]
    se = tf.keras.backend.square(y_true - y_pred)
    mse = tf.keras.backend.mean(se)
    return tf.keras.backend.sqrt(mse)


def _create_lstm(train_data: InputData):
    reg = 0.001
    shape = train_data.features.shape[-1]
    window_len = train_data.features.shape[-2]

    input_layer = tf.keras.layers.Input(shape=(window_len, shape))
    conv_first = tf.keras.layers.Conv1D(filters=16,
                                        kernel_size=16,
                                        strides=1,
                                        activation='relu',
                                        padding='same',
                                        kernel_regularizer=tf.keras.regularizers.l1(reg))(input_layer)
    conv_second = tf.keras.layers.Conv1D(filters=16,
                                         kernel_size=8,
                                         strides=1,
                                         activation='relu',
                                         padding='same',
                                         kernel_regularizer=tf.keras.regularizers.l1(reg))(input_layer)
    concatenation = tf.keras.layers.Concatenate()([input_layer, conv_first, conv_second])

    percent = (train_data.target.max() - train_data.target.min()) / 100
    noise = tf.keras.layers.GaussianNoise(percent)(concatenation)

    lstm_first_layer = tf.keras.layers.LSTM(64, return_sequences=True,
                                            kernel_regularizer=tf.keras.regularizers.l1(reg))(noise)
    lstm_second_layer = tf.keras.layers.LSTM(128, return_sequences=True,
                                             kernel_regularizer=tf.keras.regularizers.l1(reg))(lstm_first_layer)

    concatenation = tf.keras.layers.Concatenate()([concatenation, lstm_second_layer])
    output_layer = tf.keras.layers.TimeDistributed(
        tf.keras.Sequential([
            tf.keras.layers.Dropout(0.25),
            tf.keras.layers.Dense(train_data.target.shape[-1],
                                  kernel_regularizer=tf.keras.regularizers.l1(reg))
        ]))(concatenation)

    return tf.keras.Model(inputs=input_layer, outputs=output_layer)


# TODO move hyperparameters to params
def fit_lstm(train_data: InputData, epochs: int = 1):
    global forecast_length

    train_data_3d = _lagged_data_to_3d(train_data)

    ts_length = train_data_3d.features.shape[0]
    # train_data_3d.task.task_params.
    model = _create_lstm(train_data_3d)

    forecast_length = train_data_3d.task.task_params.forecast_length

    model.compile(tf.keras.optimizers.Adam(lr=0.02), loss='mse', metrics=[_rmse_only_last])

    percent = 5 * (train_data_3d.target.max() - train_data_3d.target.min()) / 100
    model.fit(train_data_3d.features, train_data_3d.target, epochs=epochs,
              callbacks=[
                  tf.keras.callbacks.EarlyStopping(monitor='loss', min_delta=percent, patience=5),
                  tf.keras.callbacks.ReduceLROnPlateau(
                      monitor='_rmse_only_last', factor=0.2, patience=2, min_delta=0.1, verbose=False),
                  tf.keras.callbacks.TensorBoard(update_freq=ts_length // 10)
              ], verbose=0)

    return model


def predict_lstm(trained_model, predict_data: InputData) -> OutputData:
    window_len, prediction_len = extract_task_param(predict_data.task)

    predict_data_3d = _lagged_data_to_3d(predict_data)

    pred = trained_model.predict(predict_data_3d.features)
    return pred[:, -prediction_len:, 0]


def _lagged_data_to_3d(input_data: InputData) -> InputData:
    transformed_data = copy(input_data)

    # TODO separate proprocessing for exog features
    transformed_data.features = np.asarray(transformed_data.features)[:, :, np.newaxis]
    transformed_data.target = np.asarray(transformed_data.target)[:, :, np.newaxis]

    return transformed_data



import numpy as np
from scipy import signal
from statsmodels.tsa.seasonal import seasonal_decompose

from fedot.core.models.data import InputData


def split_ts_to_components(trained_model, predict_data: InputData):
    """
    :param trained_model: in this case it is the value of period obtained during fitting
    :param predict_data: dataset that contains time series in target
    :return: extracted trend
    """

    target = predict_data.target
    period = trained_model

    period = _estimate_max_possible_period(period, target)

    decomposed_target = seasonal_decompose(target, period=period, extrapolate_trend='freq')
    trend = decomposed_target.trend
    residual = predict_data.target - trend

    return trend, residual


def _estimate_max_possible_period(current_period: int, target):
    # TODO implement better decomposition, now it is workaround for 'x must have 2 complete cycles' error
    if current_period * 2 >= len(target):
        current_period = round(len(target) / 2)
    return current_period


def merge_component_with_exog(component, predict_data: InputData):
    features_same_as_target = np.array_equal(predict_data.target, predict_data.features)
    if predict_data.features is not None and not features_same_as_target:
        component = component[:, None]
        component = np.concatenate((component, predict_data.features), axis=1)
    return component


def estimate_period(variable):
    analyse_ratio = 10
    f, pxx_den = signal.welch(variable, fs=1, scaling='spectrum',
                              nfft=int(len(variable) / analyse_ratio),
                              nperseg=int(len(variable) / analyse_ratio))
    period = int(1 / f[np.argmax(pxx_den)])
    return period



import os
from datetime import timedelta

from fedot.core.repository.model_types_repository import ModelTypesRepository
from examples.chain_from_automl import run_chain_from_automl
from examples.tpot_vs_fedot import run_tpot_vs_fedot_example


def test_chain_from_automl_example():
    test_file_path = str(os.path.dirname(__file__))
    experimental_repo_path = os.path.join(test_file_path,
                                          '../../fedot/core/repository/data/model_repository_with_automl.json')
    with ModelTypesRepository(experimental_repo_path) as _:
        file_path_train = os.path.join(test_file_path, 'data/simple_classification.csv')
        file_path_test = file_path_train

        auc = run_chain_from_automl(file_path_train, file_path_test, max_run_time=timedelta(seconds=1))

    assert auc > 0.5


def test_tpot_vs_fedot_example():
    test_file_path = str(os.path.dirname(__file__))
    file_path_train = os.path.join(test_file_path, 'data/simple_classification.csv')
    file_path_test = file_path_train

    auc = run_tpot_vs_fedot_example(file_path_train, file_path_test)
    assert auc > 0.5



import numpy as np
from sklearn.metrics import roc_auc_score as roc_auc

from cases.data.data_utils import get_scoring_case_data_paths
from fedot.core.composer.chain import Chain
from fedot.core.composer.node import PrimaryNode, SecondaryNode
from fedot.core.models.data import InputData


def get_case_train_test_data():
    train_file_path, test_file_path = get_scoring_case_data_paths()

    train_data = InputData.from_csv(train_file_path)
    test_data = InputData.from_csv(test_file_path)
    return train_data, test_data


def get_simple_chain():
    first = PrimaryNode(model_type='xgboost')
    second = PrimaryNode(model_type='knn')
    final = SecondaryNode(model_type='logit',
                          nodes_from=[first, second])

    chain = Chain(final)

    return chain


def chain_tuning(nodes_to_tune: str, chain: Chain, train_data: InputData,
                 test_data: InputData, local_iter: int,
                 tuner_iter_num: int = 50) -> (float, list):
    several_iter_scores_test = []

    if nodes_to_tune == 'primary':
        print('primary_node_tuning')
        chain_tune_strategy = chain.fine_tune_primary_nodes
    elif nodes_to_tune == 'root':
        print('root_node_tuning')
        chain_tune_strategy = chain.fine_tune_all_nodes
    else:
        raise ValueError(f'Invalid type of nodes. Nodes must be primary or root')

    for iteration in range(local_iter):
        print(f'current local iteration {iteration}')

        # Chain tuning
        chain_tune_strategy(train_data, iterations=tuner_iter_num)

        # After tuning prediction
        chain.fit(train_data)
        after_tuning_predicted = chain.predict(test_data)

        # Metrics
        aft_tun_roc_auc = roc_auc(y_true=test_data.target,
                                  y_score=after_tuning_predicted.predict)
        several_iter_scores_test.append(aft_tun_roc_auc)

    return float(np.mean(several_iter_scores_test)), several_iter_scores_test


if __name__ == '__main__':
    train_file_path, test_file_path = get_scoring_case_data_paths()

    train_data, test_data = get_case_train_test_data()

    # Chain composition
    chain = get_simple_chain()

    # Before tuning prediction
    chain.fit(train_data, use_cache=False)
    before_tuning_predicted = chain.predict(test_data)
    bfr_tun_roc_auc = roc_auc(y_true=test_data.target,
                              y_score=before_tuning_predicted.predict)

    local_iter = 5
    # Chain tuning
    after_tune_roc_auc, several_iter_scores_test = chain_tuning(nodes_to_tune='primary',
                                                                chain=chain,
                                                                train_data=train_data,
                                                                test_data=test_data,
                                                                local_iter=local_iter)

    print(f'Several test scores {several_iter_scores_test}')
    print(f'Mean test score over {local_iter} iterations: {after_tune_roc_auc}')
    print(round(bfr_tun_roc_auc, 3))
    print(round(after_tune_roc_auc, 3))



from typing import Callable


class SequenceIterator(object):
    """
    The value of this iterator changes according to specified sequence. Iterator is used in parameter-free evolutionary
    scheme for population size control.
    :param sequence_func: the function for sequence generation
    :param start_value: start value of sequence (if start_value doesn't match any value in the sequence, then iterator
    find closest value in sequence)
    :param max_sequence_value: maximal value in sequence
    :param min_sequence_value: minimal value in sequence
    """

    def __init__(self, sequence_func: Callable, start_value: int = None, max_sequence_value: int = None,
                 min_sequence_value: int = None):
        self.sequence_func = sequence_func
        self.archive = {}
        self.start_value = start_value
        self.index = self.get_sequence_index(self.start_value) - 1 if start_value is not None else - 1
        self.max_sequence_value = max_sequence_value
        self.min_sequence_value = min_sequence_value

    def has_prev(self) -> bool:
        if self.index > 0:
            if self.min_sequence_value is not None:
                has = self.sequence_item_calculation(self.index - 1) >= self.min_sequence_value
            else:
                has = True
        else:
            has = False
        return has

    def has_next(self):
        if self.max_sequence_value:
            has = self.max_sequence_value >= self.sequence_item_calculation(self.index + 1)
        else:
            has = True
        return has

    def sequence_item_calculation(self, index: int = None):
        index = self.index if index is None else index
        if index not in list(self.archive):
            result = self.sequence_func(index)
            self.archive[index] = result
        else:
            result = self.archive[index]
        return result

    def get_sequence_index(self, value: int) -> int:
        number = 0
        sequence_value = self.sequence_item_calculation(number)
        while sequence_value < value:
            number += 1
            sequence_value = self.sequence_item_calculation(number)
        return number

    def next(self):
        try:
            self.index += 1
            if self.min_sequence_value is not None:
                if self.sequence_item_calculation(self.index) < self.min_sequence_value:
                    self.index = self.get_sequence_index(self.min_sequence_value)
            result = self.sequence_item_calculation()
        except IndexError:
            raise StopIteration()
        return result

    def prev(self):
        self.index -= 1
        if self.index < 0:
            raise StopIteration()
        return self.sequence_item_calculation()

    def __iter__(self):
        return self


def fibonacci_sequence(n: int) -> int:
    a = 0
    b = 1
    for __ in range(n):
        a, b = b, a + b
    return a



import matplotlib.pyplot as plt
import numpy as np

from fedot.core.composer.node import PrimaryNode
from fedot.core.composer.ts_chain import TsForecastingChain
from fedot.utilities.ts_gapfilling import SimpleGapFiller, ModelGapFiller


def generate_synthetic_data(length: int = 2200, periods: int = 5):
    """
    The function generates a synthetic one-dimensional array without omissions

    :param length: the length of the array
    :param periods: the number of periods in the sine wave

    :return synthetic_data: an array without gaps
    """

    sinusoidal_data = np.linspace(-periods * np.pi, periods * np.pi, length)
    sinusoidal_data = np.sin(sinusoidal_data)
    random_noise = np.random.normal(loc=0.0, scale=0.1, size=length)

    # Combining a sine wave and random noise
    synthetic_data = sinusoidal_data + random_noise
    return synthetic_data


def generate_gaps(array_without_gaps, gap_dict, gap_value):
    """
    Function for generating gaps with predefined length in the desired indices
    of an one-dimensional array

    :param array_without_gaps: an array without gaps
    :param gap_dict: a dictionary with omissions, where the key is the index in
    the time series from which the gap will begin. The key value is the length
    of the gap (elements). -1 in the value means that a skip is generated until
    the end of the array
    :param gap_value: value indicating a gap in the array

    :return: one-dimensional array with omissions
    """

    array_with_gaps = np.copy(array_without_gaps)

    keys = list(gap_dict.keys())
    for key in keys:
        gap_size = gap_dict.get(key)
        if gap_size == -1:
            # Generating a gap to the end of an array
            array_with_gaps[key:] = gap_value
        else:
            array_with_gaps[key:(key+gap_size)] = gap_value

    return array_with_gaps


def get_array_with_gaps(gap_dict=None, gap_value: float = -100.0):
    """
    Function for generating synthetic data and gaps in it with predefined length
    and location

    :param gap_dict: a dictionary with omissions, where the key is the index in
    the time series from which the gap will begin. The key value is the length
    of the gap (elements). -1 in the value means that a skip is generated until
    the end of the array
    :param gap_value: value indicating a gap in the array

    :return array_with_gaps: an array with gaps
    :return real_values: an array with actual values in gaps
    """

    real_values = generate_synthetic_data()

    if gap_dict is None:
        gap_dict = {850: 100,
                    1400: 150}
    array_with_gaps = generate_gaps(array_without_gaps=real_values,
                                    gap_dict=gap_dict,
                                    gap_value=gap_value)

    return array_with_gaps, real_values


def run_gapfilling_example():
    """
    This function runs an example of filling in gaps in synthetic data

    :return arrays_dict: dictionary with 4 keys ('ridge', 'local_poly',
    'batch_poly', 'linear') that can be used to get arrays without gaps
    :return gap_data: an array with gaps
    :return real_data: an array with actual values in gaps
    """

    # Get synthetic time series
    gap_data, real_data = get_array_with_gaps()

    # Filling in gaps using chain from FEDOT
    ridge_chain = TsForecastingChain(PrimaryNode('ridge'))
    ridge_gapfiller = ModelGapFiller(gap_value=-100.0,
                                     chain=ridge_chain,
                                     max_window_size=150)
    without_gap_arr_ridge = \
        ridge_gapfiller.forward_inverse_filling(gap_data)

    # Filling in gaps using simple methods such as polynomial approximation
    simple_gapfill = SimpleGapFiller(gap_value=-100.0)
    without_gap_local_poly = \
        simple_gapfill.local_poly_approximation(gap_data, 4, 150)

    without_gap_batch_poly = \
        simple_gapfill.batch_poly_approximation(gap_data, 4, 150)

    without_gap_linear = \
        simple_gapfill.linear_interpolation(gap_data)

    arrays_dict = {'ridge': without_gap_arr_ridge,
                   'local_poly': without_gap_local_poly,
                   'batch_poly': without_gap_batch_poly,
                   'linear': without_gap_linear}
    return arrays_dict, gap_data, real_data


def plot_results(arrays_dict, gap_data):
    """
    Plot predictions of models

    :param arrays_dict: dictionary with 4 keys ('ridge', 'local_poly',
    'batch_poly', 'linear') that can be used to get arrays without gaps
    :param gap_data: an array with gaps
    """

    gap_ids = np.ravel(np.argwhere(gap_data == -100.0))
    masked_array = np.ma.masked_where(gap_data == -100.0, gap_data)

    plt.plot(arrays_dict.get('local_poly'), c='orange',
             alpha=0.5, label='Local polynomial approximation')
    plt.plot(arrays_dict.get('batch_poly'), c='green',
             alpha=0.5, label='Batch polynomial approximation')
    plt.plot(arrays_dict.get('linear'), c='black',
             alpha=0.5, label='Linear interpolation')
    plt.plot(arrays_dict.get('ridge'), c='red',
             alpha=0.5, label='Inverse ridge')
    plt.plot(masked_array, c='blue', alpha=1.0, label='Actual values')
    plt.xlim(gap_ids[0] - 100, gap_ids[-1] + 100)
    plt.legend(loc='upper center')
    plt.grid()
    plt.show()


# Example of applying the algorithm
if __name__ == '__main__':
    arrays_dict, gap_data, _ = run_gapfilling_example()

    # Make the plots with predictions of different models
    plot_results(arrays_dict, gap_data)



import datetime
import os

import matplotlib.pyplot as plt
from sklearn.metrics import mean_squared_error as mse

from fedot.core.composer.chain import Chain
from fedot.core.composer.gp_composer.fixed_structure_composer import FixedStructureComposerBuilder
from fedot.core.composer.gp_composer.gp_composer import GPComposerRequirements
from fedot.core.composer.node import PrimaryNode, SecondaryNode
from fedot.core.composer.visualisation import ComposerVisualiser
from fedot.core.models.data import InputData, OutputData
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.quality_metrics_repository import MetricsRepository, RegressionMetricsEnum
from fedot.core.repository.tasks import Task, TaskTypesEnum, TsForecastingParams
from fedot.core.utils import project_root


def get_composite_lstm_chain():
    chain = Chain()
    node_trend = PrimaryNode('trend_data_model')
    node_trend.labels = ["fixed"]
    node_lstm_trend = SecondaryNode('linear', nodes_from=[node_trend])
    node_trend.labels = ["fixed"]
    node_residual = PrimaryNode('residual_data_model')
    node_ridge_residual = SecondaryNode('linear', nodes_from=[node_residual])

    node_final = SecondaryNode('linear',
                               nodes_from=[node_ridge_residual, node_lstm_trend])
    node_final.labels = ["fixed"]
    chain.add_node(node_final)
    return chain


def calculate_validation_metric(pred: OutputData, valid: InputData,
                                name: str, is_visualise: bool) -> float:
    window_size = valid.task.task_params.max_window_size
    forecast_length = valid.task.task_params.forecast_length

    # skip initial part of time series
    predicted = pred.predict[window_size + forecast_length:]
    real = valid.target[window_size + forecast_length:]

    # plot results
    if is_visualise:
        compare_plot(predicted, real,
                     forecast_length=forecast_length,
                     model_name=name)

    # the quality assessment for the simulation results
    rmse = mse(y_true=real, y_pred=predicted, squared=False)

    return rmse


def compare_plot(predicted, real, forecast_length, model_name):
    plt.clf()
    _, ax = plt.subplots()
    plt.plot(real, linewidth=1, label="Observed", alpha=0.4)
    plt.plot(predicted, linewidth=1, label="Predicted", alpha=0.6)
    ax.legend()
    plt.xlabel('Time, h')
    plt.ylabel('SSH, cm')
    plt.title(f'Sea surface height forecast for {forecast_length} hours with {model_name}')
    plt.show()


def run_metocean_forecasting_problem(train_file_path, test_file_path, forecast_length=1,
                                     max_window_size=64, with_visualisation=True):
    # specify the task to solve
    task_to_solve = Task(TaskTypesEnum.ts_forecasting,
                         TsForecastingParams(forecast_length=forecast_length,
                                             max_window_size=max_window_size))

    full_path_train = os.path.join(str(project_root()), train_file_path)
    dataset_to_train = InputData.from_csv(
        full_path_train, task=task_to_solve, data_type=DataTypesEnum.ts)

    # a dataset for a final validation of the composed model
    full_path_test = os.path.join(str(project_root()), test_file_path)
    dataset_to_validate = InputData.from_csv(
        full_path_test, task=task_to_solve, data_type=DataTypesEnum.ts)

    metric_function = MetricsRepository().metric_by_id(RegressionMetricsEnum.RMSE)

    ref_chain = get_composite_lstm_chain()

    available_model_types_primary = ['trend_data_model',
                                     'residual_data_model']

    available_model_types_secondary = ['rfr', 'linear',
                                       'ridge', 'lasso']

    composer_requirements = GPComposerRequirements(
        primary=available_model_types_primary,
        secondary=available_model_types_secondary, max_arity=2,
        max_depth=4, pop_size=10, num_of_generations=10,
        crossover_prob=0, mutation_prob=0.8,
        max_lead_time=datetime.timedelta(minutes=20))

    builder = FixedStructureComposerBuilder(task=task_to_solve).with_requirements(composer_requirements).with_metrics(
        metric_function).with_initial_chain(ref_chain)
    composer = builder.build()

    chain = composer.compose_chain(data=dataset_to_train,
                                   is_visualise=False)

    if with_visualisation:
        ComposerVisualiser.visualise(chain)

    chain.fit(input_data=dataset_to_train, verbose=False)
    rmse_on_valid = calculate_validation_metric(
        chain.predict(dataset_to_validate), dataset_to_validate,
        f'full-composite_{forecast_length}',
        is_visualise=with_visualisation)

    print(f'RMSE composite: {rmse_on_valid}')

    return rmse_on_valid


if __name__ == '__main__':
    # the dataset was obtained from NEMO model simulation for sea surface height

    # a dataset that will be used as a train and test set during composition
    file_path_train = 'cases/data/metocean/metocean_data_train.csv'
    full_path_train = os.path.join(str(project_root()), file_path_train)

    # a dataset for a final validation of the composed model
    file_path_test = 'cases/data/metocean/metocean_data_test.csv'
    full_path_test = os.path.join(str(project_root()), file_path_test)

    run_metocean_forecasting_problem(full_path_train, full_path_test, forecast_length=1)



import sys
from abc import abstractmethod

import numpy as np
from sklearn.metrics import f1_score, mean_squared_error, roc_auc_score

from fedot.core.composer.chain import Chain
from fedot.core.models.data import InputData, OutputData


def from_maximised_metric(metric_func):
    def wrapper(*args, **kwargs):
        return -metric_func(*args, **kwargs)

    return wrapper


class Metric:
    @classmethod
    @abstractmethod
    def get_value(cls, chain: Chain, reference_data: InputData) -> float:
        raise NotImplementedError()


class QualityMetric:
    max_penalty_part = 0.01
    default_value = None

    @classmethod
    def get_value(cls, chain: Chain, reference_data: InputData) -> float:
        metric = cls.default_value
        try:
            results = chain.predict(reference_data)
            metric = cls.metric(reference_data, results)
        except Exception as ex:
            print(f'Metric evaluation error: {ex}')
        return metric

    @classmethod
    def get_value_with_penalty(cls, chain: Chain, reference_data: InputData) -> float:
        quality_metric = cls.get_value(chain, reference_data)
        structural_metric = StructuralComplexityMetric.get_value(chain)

        penalty = abs(structural_metric * quality_metric * cls.max_penalty_part)
        metric_with_penalty = (quality_metric +
                               min(penalty, abs(quality_metric * cls.max_penalty_part)))
        return metric_with_penalty

    @staticmethod
    @abstractmethod
    def metric(reference: InputData, predicted: OutputData) -> float:
        raise NotImplementedError()


class RmseMetric(QualityMetric):
    default_value = sys.maxsize

    @staticmethod
    def metric(reference: InputData, predicted: OutputData) -> float:
        return mean_squared_error(y_true=reference.target,
                                  y_pred=predicted.predict)


class F1Metric(QualityMetric):
    default_value = 0

    @staticmethod
    @from_maximised_metric
    def metric(reference: InputData, predicted: OutputData) -> float:
        bound = np.mean(predicted.predict)
        predicted_labels = [1 if x >= bound else 0 for x in predicted.predict]
        return f1_score(y_true=reference.target, y_pred=predicted_labels)


class MaeMetric(QualityMetric):
    default_value = sys.maxsize

    @staticmethod
    def metric(reference: InputData, predicted: OutputData) -> float:
        return mean_squared_error(y_true=reference.target, y_pred=predicted.predict)


class RocAucMetric(QualityMetric):
    default_value = 0.5

    @staticmethod
    @from_maximised_metric
    def metric(reference: InputData, predicted: OutputData) -> float:

        n_classes = reference.num_classes
        if n_classes > 2:
            additional_params = {'multi_class': 'ovo', 'average': 'macro'}
        else:
            additional_params = {}

        score = round(roc_auc_score(y_score=predicted.predict,
                                    y_true=reference.target,
                                    **additional_params), 3)
        return score


class StructuralComplexityMetric(Metric):
    @classmethod
    def get_value(cls, chain: Chain, **args) -> float:
        norm_constant = 30
        return (chain.depth ** 2 + chain.length) / norm_constant


class NodeNum(Metric):
    @classmethod
    def get_value(cls, chain: Chain, **args) -> float:
        norm_constant = 10
        return chain.length / norm_constant



from copy import deepcopy
from random import choice, random
from typing import Any, List

from fedot.core.composer.constraint import constraint_function
from fedot.core.composer.optimisers.gp_operators import \
    (equivalent_subtree, node_depth,
     nodes_from_height, replace_subtrees)
from fedot.core.utils import ComparableEnum as Enum


class CrossoverTypesEnum(Enum):
    subtree = 'subtree'
    one_point = "one_point"
    none = 'none'


def crossover(types: List[CrossoverTypesEnum], chain_first: Any, chain_second: Any, max_depth: int,
              crossover_prob: float = 0.8) -> Any:
    type = choice(types)
    chain_first_copy = deepcopy(chain_first)
    chain_second_copy = deepcopy(chain_second)
    try:
        if chain_first is chain_second or random() > crossover_prob or type == CrossoverTypesEnum.none:
            return [chain_first_copy, chain_second_copy]
        if type in crossover_by_type.keys():
            is_correct = False
            while not is_correct:
                is_correct_chains = []
                new_chains = crossover_by_type[type](chain_first_copy, chain_second_copy, max_depth)
                for new_chain in new_chains:
                    is_correct_chains.append(constraint_function(new_chain))
                is_correct = all(is_correct_chains)
            return new_chains
        else:
            raise ValueError(f'Required crossover not found: {type}')
    except Exception as ex:
        print(f'Crossover ex: {ex}')
        return chain_first_copy, chain_second_copy


def subtree_crossover(chain_first: Any, chain_second: Any, max_depth: int) -> Any:
    """Performed by the replacement of random subtree
    in first selected parent to random subtree from the second parent"""
    random_layer_in_chain_first = choice(range(chain_first.depth))
    min_second_layer = 1 if random_layer_in_chain_first == 0 else 0
    random_layer_in_chain_second = choice(range(min_second_layer, chain_second.depth))

    node_from_chain_first = choice(nodes_from_height(chain_first, random_layer_in_chain_first))
    node_from_chain_second = choice(nodes_from_height(chain_second, random_layer_in_chain_second))

    replace_subtrees(chain_first, chain_second, node_from_chain_first, node_from_chain_second,
                     random_layer_in_chain_first, random_layer_in_chain_second, max_depth)

    return chain_first, chain_second


def one_point_crossover(chain_first: Any, chain_second: Any, max_depth: int) -> Any:
    """Finds common structural parts between two trees, and after that randomly
    chooses the location of nodes, subtrees of which will be swapped"""
    pairs_of_nodes = equivalent_subtree(chain_first, chain_second)
    if pairs_of_nodes:
        node_from_chain_first, node_from_chain_second = choice(pairs_of_nodes)

        layer_in_chain_first = node_depth(chain_first.root_node) - node_depth(node_from_chain_first)
        layer_in_chain_second = node_depth(chain_second.root_node) - node_depth(node_from_chain_second)

        replace_subtrees(chain_first, chain_second, node_from_chain_first, node_from_chain_second,
                         layer_in_chain_first, layer_in_chain_second, max_depth)
    return chain_first, chain_second


crossover_by_type = {
    CrossoverTypesEnum.subtree: subtree_crossover,
    CrossoverTypesEnum.one_point: one_point_crossover,
}



from typing import Optional

import numpy as np
from sklearn.decomposition import PCA

from fedot.core.algorithms.time_series.scale import estimate_period, split_ts_to_components
from fedot.core.models.data import InputData
from fedot.core.models.evaluation.evaluation import EvaluationStrategy

DEFAULT_EXPLAINED_VARIANCE_THR = 0.9
DEFAULT_MIN_EXPLAINED_VARIANCE = 0.01


def get_data(trained_model, predict_data: InputData):
    return predict_data.features


def get_difference(trained_model, predict_data: InputData):
    number_of_inputs = predict_data.features.shape[1]
    if number_of_inputs != 1:
        raise ValueError('Too many inputs for the differential model')
    return predict_data.features[:, 0] - predict_data.target


def fit_decomposition(train_data: InputData, params: Optional[dict]):
    target = train_data.target
    period = estimate_period(target)
    return period


def get_residual(trained_model, predict_data: InputData):
    _, target_residual = split_ts_to_components(trained_model, predict_data)
    return target_residual


def get_trend(trained_model, predict_data: InputData):
    target_trend, _ = split_ts_to_components(trained_model, predict_data)
    return target_trend


def fit_pca(train_data: InputData, params: Optional[dict]):
    if not params:
        pca = PCA(svd_solver='randomized', iterated_power='auto')
    else:
        pca_params = {k: params[k] for k in ['svd_solver', 'iterated_power']}
        pca = PCA(**pca_params)

    pca.fit(train_data.features)

    cum_variance = np.cumsum(pca.explained_variance_ratio_)

    explained_variance_thr = DEFAULT_EXPLAINED_VARIANCE_THR
    min_explained_variance = DEFAULT_MIN_EXPLAINED_VARIANCE

    if params:
        explained_variance_thr = params.get('dim_reduction_expl_thr', explained_variance_thr)
        min_explained_variance = params.get('dim_reduction_min_expl', min_explained_variance)

    components_before_thr = np.where(cum_variance > explained_variance_thr)[0]

    last_ind_cum = min(components_before_thr) if len(components_before_thr) > 0 else 1

    significant_components = \
        np.where(pca.explained_variance_ratio_ < min_explained_variance)[0]

    last_ind = min(significant_components) if len(significant_components) > 0 else 1

    pca.last_component_ind = min(last_ind, last_ind_cum)

    return pca


def predict_pca(pca_model, predict_data: InputData):
    return pca_model.transform(predict_data.features)[:, :(pca_model.last_component_ind + 1)]


class DataModellingStrategy(EvaluationStrategy):
    _model_functions_by_type = {
        'direct_data_model': (None, get_data),
        'trend_data_model': (fit_decomposition, get_trend),
        'residual_data_model': (fit_decomposition, get_residual),
        'pca_data_model': (fit_pca, predict_pca)
    }

    def __init__(self, model_type: str, params: Optional[dict] = None):
        self._model_specific_fit = self._model_functions_by_type[model_type][0]
        self._model_specific_predict = self._model_functions_by_type[model_type][1]
        super().__init__(model_type, params)

    def fit(self, train_data: InputData):
        if not self._model_specific_fit:
            return None
        else:
            return self._model_specific_fit(train_data, self.params_for_fit)

    def predict(self, trained_model, predict_data: InputData):
        return self._model_specific_predict(trained_model, predict_data)

    def fit_tuned(self, **args):
        return None, None



import os
from pathlib import Path

from fedot.core.utils import default_fedot_data_dir


def test_default_fedot_data_dir():
    default_fedot_data_dir()
    assert 'Fedot' in os.listdir(str(Path.home()))



from random import seed

import matplotlib.pyplot as plt
import numpy as np
from sklearn.metrics import roc_auc_score as roc_auc

from fedot.core.composer.chain import Chain
from fedot.core.composer.node import PrimaryNode, SecondaryNode
from fedot.core.models.data import train_test_data_setup
from fedot.utilities.synthetic.chain import separately_fit_chain
from fedot.utilities.synthetic.data import (
    classification_dataset, gauss_quantiles_dataset
)
from fedot.utilities.synthetic.data_benchmark import synthetic_benchmark_dataset


def data_generator_example():
    samples_total, features_amount = 100, 10
    classes = 2
    options = {
        'informative': 8,
        'redundant': 1,
        'repeated': 1,
        'clusters_per_class': 1
    }
    features, target = classification_dataset(samples_total, features_amount, classes,
                                              features_options=options,
                                              noise_fraction=0.1, full_shuffle=False)

    plt.subplot(121)
    plt.title('The first two informative features, one cluster per class')
    plt.scatter(features[:, 0], features[:, 1], marker='o', c=target,
                s=25, edgecolor='k')

    features, target = gauss_quantiles_dataset(samples_total, features_amount=2, classes_amount=classes)
    plt.subplot(122)
    plt.title(f'Gaussian divided into {classes} quantiles')
    plt.scatter(features[:, 0], features[:, 1], marker='o', c=target,
                s=25, edgecolor='k')

    plt.show()


def synthetic_benchmark_composing_example():
    fitted_chain = separately_fit_chain(samples=5000, features_amount=10,
                                        classes=2)
    data = synthetic_benchmark_dataset(samples_amount=5000, features_amount=10,
                                       fitted_chain=fitted_chain)

    print(f'Synthetic features: {data.features[:10]}')
    print(f'Synthetic target: {data.target[:10]}')

    train, test = train_test_data_setup(data)
    simple_chain = two_level_chain()
    simple_chain.fit(input_data=train, use_cache=False)

    print(f'ROC score on train: {roc_value(simple_chain, train)}')
    print(f'ROC score on test {roc_value(simple_chain, test)}')


def two_level_chain():
    first = PrimaryNode(model_type='logit')
    second = PrimaryNode(model_type='knn')
    third = SecondaryNode(model_type='xgboost',
                          nodes_from=[first, second])

    chain = Chain()
    for node in [first, second, third]:
        chain.add_node(node)

    return chain


def roc_value(chain: Chain, dataset_to_validate) -> float:
    predicted = chain.predict(dataset_to_validate)
    roc_auc_value = roc_auc(y_true=dataset_to_validate.target,
                            y_score=predicted.predict)
    return roc_auc_value


if __name__ == '__main__':
    seed(15)
    np.random.seed(15)
    data_generator_example()
    synthetic_benchmark_composing_example()



import itertools
import uuid
from copy import copy
from dataclasses import dataclass
from typing import List

import numpy as np

from fedot.core.composer.chain import Chain
from fedot.core.composer.node import CachedState
from fedot.core.composer.node import Node
from fedot.core.composer.node import PrimaryNode, SecondaryNode, FittedModelCache
from fedot.core.models.model import InputData
from fedot.core.models.model import Model
from fedot.core.models.preprocessing import Normalization
from fedot.core.repository.tasks import Task, TaskTypesEnum
from fedot.utilities.synthetic.data import classification_dataset as synthetic_dataset
from fedot.utilities.synthetic.data import gauss_quantiles_dataset as gauss_quantiles
from fedot.core.repository.dataset_types import DataTypesEnum


@dataclass
class ModelTemplate:
    def __init__(self, model_type):
        self.id = str(uuid.uuid4())
        self.model_type = model_type
        self.input_shape = []
        self.output_shape = []
        self.parents = []
        self.model_instance = None
        self.fitted_model = None
        self.data_fit = None
        self.preprocessor = None

    def __eq__(self, other):
        return self.id == other.id


def chain_template_random(model_types, depth, models_per_level,
                          samples, features):
    models_by_level = []

    for level in range(depth - 1):
        selected_models = np.random.choice(model_types, models_per_level[level])
        templates = [ModelTemplate(model_type=model_type) for model_type in selected_models]
        models_by_level.append(templates)

    root_model = np.random.choice(model_types)
    models_by_level.append([ModelTemplate(model_type=root_model)])

    models_by_level = with_random_links(models_by_level)

    features_shape = [samples, features]
    # TODO: change target_shape later if non-classification problems will be used
    target_shape = [samples, 1]
    models_by_level = with_calculated_shapes(models_by_level,
                                             features_shape=features_shape,
                                             target_shape=target_shape)
    return models_by_level


def chain_template_balanced_tree(model_types, depth, models_per_level,
                                 samples, features):
    models_by_level = []

    for level in range(depth - 1):
        selected_models = np.random.choice(model_types, models_per_level[level])
        templates = [ModelTemplate(model_type=model_type) for model_type in selected_models]
        models_by_level.append(templates)

    assert models_per_level[-1] == 1
    root_model = np.random.choice(model_types)
    models_by_level.append([ModelTemplate(model_type=root_model)])

    models_by_level = with_balanced_tree_links(models_by_level)

    features_shape = [samples, features]
    # TODO: change target_shape later if non-classification problems will be used
    target_shape = [samples, 1]
    models_by_level = with_calculated_shapes(models_by_level,
                                             features_shape=features_shape,
                                             target_shape=target_shape)
    return models_by_level


# TODO: fix internal primary nodes appearance
def with_random_links(models_by_level):
    for current_lvl in range(len(models_by_level) - 1):
        next_lvl = current_lvl + 1
        models_on_current_lvl = models_by_level[current_lvl]

        for model in models_on_current_lvl:
            models_on_next_lvl = models_by_level[next_lvl]
            links_amount = np.random.randint(1, len(models_on_next_lvl) + 1)
            models_to_link = np.random.choice(models_by_level[next_lvl], links_amount, replace=False)
            for model_ in models_to_link:
                model_.parents.append(model)
    return models_by_level


def with_balanced_tree_links(models_by_level):
    for current_lvl in range(len(models_by_level) - 1):
        next_lvl = current_lvl + 1
        models_on_lvl = models_by_level[current_lvl]

        current_level_amount = len(models_by_level[current_lvl])
        next_level_amount = len(models_by_level[next_lvl])
        assert current_level_amount >= next_level_amount

        models_per_group = current_level_amount // next_level_amount

        current_group_idx = 0
        for model_idx in range(current_level_amount):
            if (model_idx % models_per_group) == 0 and model_idx != 0 and current_group_idx < next_level_amount - 1:
                current_group_idx += 1
            current_model = models_on_lvl[model_idx]
            current_group_parent = models_by_level[next_lvl][current_group_idx]

            current_group_parent.parents.append(current_model)

    return models_by_level


def with_calculated_shapes(models_by_level, features_shape, target_shape):
    samples_idx, classes_idx = 0, 1
    samples, classes = features_shape[samples_idx], target_shape[classes_idx]

    # Fill the first level
    for model in models_by_level[0]:
        model.input_shape = features_shape
        model.output_shape = target_shape

    for level in range(1, len(models_by_level)):
        for model in models_by_level[level]:
            if len(model.parents) == 0:
                input_features = features_shape[classes_idx]
            else:
                input_features = sum([parent.output_shape[classes_idx] for parent in model.parents])

            model.input_shape = [samples, input_features]

            model.output_shape = target_shape

    return models_by_level


def show_chain_template(models_by_level):
    for level in range(len(models_by_level)):
        print(f'Level = {level}')
        for model in models_by_level[level]:
            print(f'{model.model_type}, input = {model.input_shape}, output = {model.output_shape}')


# TODO: refactor skip_fit logic
def fit_template(chain_template, classes, with_gaussian=False, skip_fit=False):
    templates_by_models = []
    for model_template in itertools.chain.from_iterable(chain_template):
        model_instance = Model(model_type=model_template.model_type)
        model_template.model_instance = model_instance
        templates_by_models.append((model_template, model_instance))
    if skip_fit:
        return

    for template, instance in templates_by_models:
        samples, features_amount = template.input_shape

        if with_gaussian:
            features, target = gauss_quantiles(samples_amount=samples,
                                               features_amount=features_amount,
                                               classes_amount=classes)
        else:
            options = {
                'informative': features_amount,
                'redundant': 0,
                'repeated': 0,
                'clusters_per_class': 1
            }
            features, target = synthetic_dataset(samples_amount=samples,
                                                 features_amount=features_amount,
                                                 classes_amount=classes,
                                                 features_options=options)
        target = np.expand_dims(target, axis=1)
        data_train = InputData(idx=np.arange(0, samples),
                               features=features, target=target,
                               data_type=DataTypesEnum.table,
                               task=Task(TaskTypesEnum.classification))

        preproc_data = copy(data_train)
        preprocessor = Normalization().fit(preproc_data.features)
        preproc_data.features = preprocessor.apply(preproc_data.features)
        print(f'Fit {instance}')
        fitted_model, predictions = instance.fit(data=preproc_data)

        template.fitted_model = fitted_model
        template.data_fit = preproc_data
        template.preprocessor = preprocessor


def real_chain(chain_template, with_cache=True):
    nodes_by_templates = []
    for level in range(0, len(chain_template)):
        for template in chain_template[level]:
            if len(template.parents) == 0:
                node = PrimaryNode(model_type=template.model_type)
            else:
                node = SecondaryNode(nodes_from=real_parents(nodes_by_templates,
                                                             template),
                                     model_type=template.model_type)
            node.model = template.model_instance
            if with_cache:
                cache = FittedModelCache(related_node=node)
                cache.append(CachedState(preprocessor=template.preprocessor, model=template.fitted_model))
                node.cache = cache
            nodes_by_templates.append((node, template))

    chain = Chain()
    for node, _ in nodes_by_templates:
        chain.add_node(node)

    return chain


def real_parents(nodes_by_templates, template_child) -> List[Node]:
    parent_nodes = []
    for parent in template_child.parents:
        for node, template in nodes_by_templates:
            if template == parent:
                parent_nodes.append(node)
                break
    return parent_nodes



from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import roc_auc_score as roc_auc
from sklearn.naive_bayes import BernoulliNB
from sklearn.pipeline import make_pipeline
from tpot.builtins import StackingEstimator
from tpot.export_utils import set_param_recursive

from fedot.core.composer.chain import Chain
from fedot.core.composer.node import PrimaryNode, SecondaryNode
from fedot.core.models.data import InputData


def run_tpot_vs_fedot_example(train_file_path: str, test_file_path: str):
    train_data = InputData.from_csv(train_file_path)
    test_data = InputData.from_csv(test_file_path)

    training_features = train_data.features
    testing_features = test_data.features
    training_target = train_data.target
    testing_target = test_data.target

    # Average CV score on the training set was: 0.93755
    exported_pipeline = make_pipeline(
        StackingEstimator(estimator=BernoulliNB()),
        RandomForestClassifier()
    )
    # Fix random state for all the steps in exported pipeline
    set_param_recursive(exported_pipeline.steps, 'random_state', 1)

    exported_pipeline.fit(training_features, training_target)
    results = exported_pipeline.predict_proba(testing_features)[:, 1]

    roc_auc_value = roc_auc(y_true=testing_target,
                            y_score=results)

    print(roc_auc_value)

    chain = Chain()
    node_first = PrimaryNode('direct_data_model')
    node_second = PrimaryNode('bernb')
    node_third = SecondaryNode('rf')

    node_third.nodes_from.append(node_first)
    node_third.nodes_from.append(node_second)

    chain.add_node(node_third)

    chain.fit(train_data)
    results = chain.predict(test_data)

    roc_auc_value = roc_auc(y_true=testing_target,
                            y_score=results.predict)
    print(roc_auc_value)

    return roc_auc_value


if __name__ == '__main__':
    train_file_path = "../cases/data/scoring/scoring_train.csv"
    test_file_path = "../cases/data/scoring/scoring_test.csv"

    run_tpot_vs_fedot_example(train_file_path, test_file_path)



from datetime import timedelta
from typing import Optional

import numpy as np
from statsmodels.tsa.ar_model import AutoReg
from statsmodels.tsa.arima_model import ARIMA

from fedot.core.models.data import InputData, OutputData
from fedot.core.models.evaluation.evaluation import EvaluationStrategy
from fedot.core.models.tuning.tuners import ForecastingCustomRandomTuner


def fit_ar(train_data: InputData, params):
    return AutoReg(train_data.target, **params,
                   exog=train_data.features).fit()


def fit_arima(train_data: InputData, params):
    return ARIMA(train_data.target, **params).fit(disp=0)


def predict_ar(trained_model, predict_data: InputData) -> OutputData:
    prediction = None

    for ts_step_id, target in enumerate(predict_data.features):
        start, end = trained_model.nobs + ts_step_id, \
                     trained_model.nobs + ts_step_id + predict_data.task.task_params.forecast_length
        exog, exog_oos = None, predict_data.features[start:end]

        if trained_model.data.endog is predict_data.target:
            # if train sample used
            start, end = 0 + ts_step_id, predict_data.task.task_params.forecast_length + ts_step_id
            exog, exog_oos = (predict_data.features[start:end],
                              predict_data.features[start:end])

        prediction = trained_model.predict(start=start, end=end,
                                           exog=exog, exog_oos=exog_oos)

    return prediction


def predict_arima(trained_model, predict_data: InputData):
    prediction = []

    forecast_length = predict_data.task.task_params.forecast_length
    prediction_steps = len(predict_data.features) + 1
    for pred_step_id in range(prediction_steps):

        start, end = trained_model.nobs + pred_step_id, \
                     trained_model.nobs + pred_step_id + forecast_length - 1

        if trained_model.data.endog is predict_data.target:
            # if train sample used
            start, end = 0 + pred_step_id, forecast_length + pred_step_id

        # TODO take exog into account
        prediction_for_step = trained_model.predict(start=start, end=end)
        prediction_for_step = prediction_for_step[-forecast_length:]
        prediction.append(prediction_for_step)

    return np.stack(prediction)


class StatsModelsForecastingStrategy(EvaluationStrategy):
    __model_functions_by_types = {
        'arima': (fit_arima, predict_arima),
        'ar': (fit_ar, predict_ar)
    }

    __model_description_by_func = {
        fit_arima: 'statsmodels.tsa.arima_model.ARIMA',
        fit_ar: 'statsmodels.tsa.ar_model import AutoReg'
    }

    __default_params_by_model = {
        'arima': {'order': (2, 0, 0)},
        'ar': {'lags': (1, 2, 6, 12, 24)}
    }
    __params_range_by_model = {
        'arima': {'order': ((2, 0, 0), (5, 0, 5))},
        'ar': {'lags': (range(1, 6), range(6, 96, 6))}
    }

    def __init__(self, model_type: str, params: Optional[dict] = None):
        self._model_specific_fit, self._model_specific_predict = self._init_stats_model_functions(model_type)
        self._params_range = self.__params_range_by_model[model_type]
        self._default_params = self.__default_params_by_model[model_type]
        super().__init__(model_type, params)

    def _init_stats_model_functions(self, model_type: str):
        if model_type in self.__model_functions_by_types.keys():
            return self.__model_functions_by_types[model_type]
        else:
            raise ValueError(f'Impossible to obtain Stats strategy for {model_type}')

    def fit(self, train_data: InputData):
        stats_model = self._model_specific_fit(train_data, self._default_params)
        self.params_for_fit = self._default_params
        return stats_model

    def predict(self, trained_model, predict_data: InputData) -> OutputData:
        return self._model_specific_predict(trained_model, predict_data)

    def fit_tuned(self, train_data: InputData, iterations: int = 10,
                  max_lead_time: timedelta = timedelta(minutes=5)):
        tuned_params = ForecastingCustomRandomTuner().tune(fit=self._model_specific_fit,
                                                           predict=self._model_specific_predict,
                                                           tune_data=train_data,
                                                           params_range=self._params_range,
                                                           default_params=self._default_params,
                                                           iterations=iterations)

        stats_model = self._model_specific_fit(train_data, tuned_params)
        self.params_for_fit = tuned_params

        return stats_model, tuned_params

    @property
    def implementation_info(self) -> str:
        return self.__model_description_by_func[self._model_specific_fit]



from copy import deepcopy
import numpy as np
from typing import (Optional, List, Any, Tuple)
from fedot.core.composer.optimisers.inheritance import GeneticSchemeTypesEnum, inheritance
from fedot.core.composer.optimisers.regularization import regularized_population
from fedot.core.composer.optimisers.selection import selection
from fedot.core.composer.optimisers.gp_optimiser import GPChainOptimiserParameters, GPChainOptimiser
from fedot.core.composer.timer import CompositionTimer
from fedot.core.composer.iterator import fibonacci_sequence, SequenceIterator
from fedot.core.composer.optimisers.gp_operators import num_of_parents_in_crossover
from fedot.core.log import Log


class GPChainParameterFreeOptimiser(GPChainOptimiser):
    """
    Implementation of the parameter-free adaptive evolutionary optimiser
    (population size and genetic operators rates is changing over time).
    For details, see original paper: https://arxiv.org/abs/2001.10178
    :param initial_chain: chain which was initialized outside the optimiser
    :param requirements: composer requirements
    :param chain_generation_params: parameters for new chain generation
    :param parameters: parameters of chain optimiser
    :param max_population_size: maximum population size
    :param log: optional parameter for log object
    """

    def __init__(self, initial_chain, requirements, chain_generation_params,
                 parameters: Optional[GPChainOptimiserParameters] = None,
                 max_population_size: int = 55,
                 sequence_function=fibonacci_sequence, log: Log = None):
        super().__init__(initial_chain, requirements, chain_generation_params, parameters, log)

        if self.parameters.genetic_scheme_type != GeneticSchemeTypesEnum.parameter_free:
            self.log.error(f'Invalid genetic scheme type was changed to parameter-free . Continue.')
            self.parameters.genetic_scheme_type = GeneticSchemeTypesEnum.parameter_free

        self.sequence_function = sequence_function
        self.max_pop_size = max_population_size
        self.iterator = SequenceIterator(sequence_func=self.sequence_function, min_sequence_value=1,
                                         max_sequence_value=self.max_pop_size,
                                         start_value=self.requirements.pop_size)
        self.generation_num = 0
        self.requirements.pop_size = self.iterator.next()

    def optimise(self, objective_function, offspring_rate: float = 0.5):
        if self.population is None:
            self.population = self._make_population(self.requirements.pop_size)

        num_of_new_individuals = self.offspring_size(offspring_rate)
        self.log.info(f'pop size: {self.requirements.pop_size}, num of new inds: {num_of_new_individuals}')
        with CompositionTimer() as t:

            self.history = []

            if self.requirements.add_single_model_chains:
                best_single_model, self.requirements.primary = \
                    self._best_single_models(objective_function)

            for ind in self.population:
                ind.fitness = objective_function(ind)

            self._add_to_history(self.population)

            self.log.info(f'Best metric is {self.best_individual.fitness}')

            while not t.is_time_limit_reached(self.requirements.max_lead_time) \
                    and self.generation_num != self.requirements.num_of_generations - 1:
                self.log.info(f'Generation num: {self.generation_num}')
                self.num_of_gens_without_improvements = self.update_stagnation_counter()
                self.log.info(f'max_depth: {self.max_depth}, no improvements: {self.num_of_gens_without_improvements}')
                if self.parameters.with_auto_depth_configuration and self.generation_num != 0:
                    self.max_depth_recount()

                self.max_std = self.update_max_std()

                individuals_to_select = regularized_population(reg_type=self.parameters.regularization_type,
                                                               population=self.population,
                                                               objective_function=objective_function,
                                                               chain_class=self.chain_class)

                if num_of_new_individuals == 1 and len(self.population) == 1:
                    new_population = list(self.reproduce(self.population[0]))
                    new_population[0].fitness = objective_function(new_population[0])
                else:
                    num_of_parents = num_of_parents_in_crossover(num_of_new_individuals)

                    selected_individuals = selection(types=self.parameters.selection_types,
                                                     population=individuals_to_select,
                                                     pop_size=num_of_parents)

                    new_population = []

                    for parent_num in range(0, len(selected_individuals), 2):
                        new_population += self.reproduce(selected_individuals[parent_num],
                                                         selected_individuals[parent_num + 1])

                        new_population[parent_num].fitness = objective_function(new_population[parent_num])
                        new_population[parent_num + 1].fitness = objective_function(new_population[parent_num + 1])

                self.requirements.pop_size = self.next_population_size(new_population)
                num_of_new_individuals = self.offspring_size(offspring_rate)
                self.log.info(f'pop size: {self.requirements.pop_size}, num of new inds: {num_of_new_individuals}')

                self.prev_best = deepcopy(self.best_individual)

                self.population = inheritance(self.parameters.genetic_scheme_type, self.parameters.selection_types,
                                              self.population,
                                              new_population, self.num_of_inds_in_next_pop)

                if self.with_elitism:
                    self.population.append(self.prev_best)

                self._add_to_history(self.population)
                self.log.info(f'spent time: {round(t.minutes_from_start, 1)} min')
                self.log.info(f'Best metric is {self.best_individual.fitness}')

                self.generation_num += 1

            best = self.best_individual
            if self.requirements.add_single_model_chains and \
                    (best_single_model.fitness <= best.fitness):
                best = best_single_model
        return best, self.history

    @property
    def with_elitism(self) -> bool:
        return self.requirements.pop_size >= 10

    @property
    def current_std(self):
        if self.requirements.pop_size == 1 and self.generation_num == 0:
            std = 0
        else:
            std = np.std([ind.fitness for ind in self.population])
        return std

    def update_max_std(self):
        if self.generation_num == 0:
            std_max = self.current_std
            self.requirements.mutation_prob = 1
            self.requirements.crossover_prob = 0
        else:
            if self.max_std < self.current_std:
                std_max = self.current_std
            else:
                std_max = self.max_std
        return std_max

    def next_population_size(self, offspring: List[Any]) -> int:
        best_in_offspring = self.get_best_individual(offspring, equivalents_from_current_pop=False)
        fitness_improved = best_in_offspring.fitness < self.best_individual.fitness
        complexity_decreased = len(best_in_offspring.nodes) < len(
            self.best_individual.nodes) and best_in_offspring.fitness <= self.best_individual.fitness
        is_max_pop_size_reached = not self.iterator.has_next()
        progress_in_both_goals = fitness_improved and complexity_decreased and not is_max_pop_size_reached
        no_progress = not fitness_improved and not complexity_decreased and not is_max_pop_size_reached
        if (progress_in_both_goals and len(self.population) > 2) or no_progress:
            if progress_in_both_goals:
                if self.iterator.has_prev():
                    next_population_size = self.iterator.prev()
                else:
                    next_population_size = len(self.population)
            else:
                next_population_size = self.iterator.next()

                self.requirements.mutation_prob, self.requirements.crossover_prob = self.operators_prob_update(
                    std=float(self.current_std), max_std=float(self.max_std))

        else:
            next_population_size = len(self.population)
        return next_population_size

    def operators_prob_update(self, std: float, max_std: float) -> Tuple[float, float]:
        mutation_prob = 1 - (std / max_std) if max_std > 0 and std != max_std else 0.5
        crossover_prob = self.requirements.crossover_prob = 1 - self.requirements.mutation_prob
        return mutation_prob, crossover_prob

    def offspring_size(self, offspring_rate: float = None) -> int:
        if self.iterator.has_prev():
            num_of_new_individuals = self.iterator.prev()
            self.iterator.next()
        else:
            num_of_new_individuals = 1
        return num_of_new_individuals



from typing import Union

# imports are required for the eval
from fedot.core.repository.dataset_types import *
from fedot.core.repository.tasks import *


def read_field(source, field_name, default):
    if field_name in source.keys():
        field_value = source[field_name]
        if isinstance(field_value, str):
            return eval_field_str(field_value)
        else:
            return field_value
    else:
        return default


def eval_field_str(field_value) -> Union[List[DataTypesEnum],
                                         List[TaskTypesEnum]]:
    return eval(field_value)


def eval_strategy_str(field_value):
    namespace = field_value[0]
    exec(f'from {namespace} import {field_value[1]}')
    return eval(field_value[1])



from typing import Optional

import networkx as nx
from networkx.algorithms.cycles import simple_cycles
from networkx.algorithms.isolate import isolates

from fedot.core.composer.chain import Chain, as_nx_graph
from fedot.core.composer.node import PrimaryNode, SecondaryNode
from fedot.core.repository.tasks import Task

ERROR_PREFIX = 'Invalid chain configuration:'


def validate(chain: Chain, task: Optional[Task] = None):
    # TODO pass task to this function
    has_one_root(chain)
    has_no_cycle(chain)
    has_no_self_cycled_nodes(chain)
    has_no_isolated_nodes(chain)
    has_primary_nodes(chain)
    has_correct_model_positions(chain, task)
    return True


def has_one_root(chain: Chain):
    if chain.root_node:
        return True


def has_no_cycle(chain: Chain):
    graph, _ = as_nx_graph(chain)
    cycled = list(simple_cycles(graph))
    if len(cycled) > 0:
        raise ValueError(f'{ERROR_PREFIX} Chain has cycles')

    return True


def has_no_isolated_nodes(chain: Chain):
    graph, _ = as_nx_graph(chain)
    isolated = list(isolates(graph))
    if len(isolated) > 0 and chain.length != 1:
        raise ValueError(f'{ERROR_PREFIX} Chain has isolated nodes')
    return True


def has_primary_nodes(chain: Chain):
    if not any(node for node in chain.nodes if isinstance(node, PrimaryNode)):
        raise ValueError(f'{ERROR_PREFIX} Chain does not have primary nodes')
    return True


def has_no_self_cycled_nodes(chain: Chain):
    if any([node for node in chain.nodes if isinstance(node, SecondaryNode) and node in node.nodes_from]):
        raise ValueError(f'{ERROR_PREFIX} Chain has self-cycled nodes')
    return True


def has_no_isolated_components(chain: Chain):
    graph, _ = as_nx_graph(chain)
    ud_graph = nx.Graph()
    ud_graph.add_nodes_from(graph)
    ud_graph.add_edges_from(graph.edges)
    if not nx.is_connected(ud_graph):
        raise ValueError(f'{ERROR_PREFIX} Chain has isolated components')
    return True


def _is_data_merged(chain: Chain):
    root_node_merges_data = 'composition' in chain.root_node.model_tags
    merging_is_required = any('decomposition' in node.model_tags for node in chain.nodes)
    data_merged_or_merging_not_required = root_node_merges_data or not merging_is_required

    return data_merged_or_merging_not_required


def _is_primary_not_composition_datamodel(chain: Chain):
    is_primary_not_composition_datamodel = all(['composition' not in node.model_tags
                                                for node in chain.nodes if isinstance(node, PrimaryNode)])
    return is_primary_not_composition_datamodel


def _is_root_not_datamodel(chain: Chain):
    return 'data_model' not in chain.root_node.model_tags and \
           'decomposition' not in chain.root_node.model_tags


def has_correct_model_positions(chain: Chain, task: Optional[Task] = None):
    is_root_satisfy_task_type = True
    if task:
        is_root_satisfy_task_type = task.task_type in chain.root_node.model.acceptable_task_types

    if not _is_root_not_datamodel(chain) or \
            not _is_primary_not_composition_datamodel(chain) or \
            not _is_data_merged(chain) or \
            not is_root_satisfy_task_type:
        raise ValueError(f'{ERROR_PREFIX} Chain has incorrect models positions')

    return True



from copy import deepcopy
from typing import (Any, Callable, List, Optional)

from fedot.core.composer.constraint import constraint_function
from fedot.core.utils import ComparableEnum as Enum


class RegularizationTypesEnum(Enum):
    none = 'none'
    decremental = 'decremental'


def regularized_population(reg_type: RegularizationTypesEnum, population: List[Any],
                           objective_function: Callable,
                           chain_class: Any, size: Optional[int] = None) -> List[Any]:
    if reg_type == RegularizationTypesEnum.decremental:
        additional_inds = decremental_regularization(population, objective_function, chain_class, size)
        return population + additional_inds
    elif reg_type == RegularizationTypesEnum.none:
        return population
    else:
        raise ValueError(f'Required regularization type not found: {type}')


def decremental_regularization(population: List[Any], objective_function: Callable,
                               chain_class: Any, size: Optional[int] = None) -> List[Any]:
    size = size if size else len(population)
    additional_inds = []
    prev_nodes_ids = []
    for ind in population:
        ind_subtrees = [node for node in ind.nodes if node != ind.root_node]
        subtrees = [chain_class(deepcopy(node.ordered_subnodes_hierarchy)) for node in ind_subtrees if
                    is_fitted_subtree(node, prev_nodes_ids)]
        additional_inds += subtrees
        prev_nodes_ids += [subtree.root_node.descriptive_id for subtree in subtrees]

    additional_inds = [ind for ind in additional_inds if constraint_function(ind)]

    for additional_ind in additional_inds:
        additional_ind.fitness = objective_function(additional_ind)

    if additional_inds and len(additional_inds) > size:
        additional_inds = sorted(additional_inds, key=lambda ind: ind.fitness)[:size]

    return additional_inds


def is_fitted_subtree(node: Any, prev_nodes_ids: List[Any]) -> bool:
    return node.nodes_from and not node.descriptive_id in prev_nodes_ids and node.cache.actual_cached_state



import datetime
import random
from datetime import timedelta

import numpy as np
from sklearn.metrics import roc_auc_score as roc_auc

from fedot.core.composer.chain import Chain
from fedot.core.composer.gp_composer.gp_composer import \
    GPComposerBuilder, GPComposerRequirements
from fedot.core.composer.visualisation import ComposerVisualiser
from fedot.core.models.data import InputData
from fedot.core.repository.model_types_repository import ModelTypesRepository
from fedot.core.repository.quality_metrics_repository import \
    ClassificationMetricsEnum, MetricsRepository
from fedot.core.repository.tasks import Task, TaskTypesEnum
from fedot.core.utils import probs_to_labels
from examples.utils import create_multi_clf_examples_from_excel

random.seed(1)
np.random.seed(1)


def get_model(train_file_path: str, cur_lead_time: datetime.timedelta = timedelta(seconds=60)):
    task = Task(task_type=TaskTypesEnum.classification)
    dataset_to_compose = InputData.from_csv(train_file_path, task=task)

    # the search of the models provided by the framework
    # that can be used as nodes in a chain for the selected task
    models_repo = ModelTypesRepository()
    available_model_types, _ = models_repo.suitable_model(task_type=task.task_type)

    metric_function = MetricsRepository(). \
        metric_by_id(ClassificationMetricsEnum.ROCAUC_penalty)

    composer_requirements = GPComposerRequirements(
        primary=available_model_types, secondary=available_model_types,
        max_lead_time=cur_lead_time)

    # Create the genetic programming-based composer, that allow to find
    # the optimal structure of the composite model
    builder = GPComposerBuilder(task).with_requirements(composer_requirements).with_metrics(metric_function)
    composer = builder.build()

    # run the search of best suitable model
    chain_evo_composed = composer.compose_chain(data=dataset_to_compose, is_visualise=False)
    chain_evo_composed.fit(input_data=dataset_to_compose)

    return chain_evo_composed


def apply_model_to_data(model: Chain, data_path: str):
    df, file_path = create_multi_clf_examples_from_excel(data_path, return_df=True)
    dataset_to_apply = InputData.from_csv(file_path, target_column=None)
    evo_predicted = model.predict(dataset_to_apply)
    df['forecast'] = probs_to_labels(evo_predicted.predict)
    return df


def validate_model_quality(model: Chain, data_path: str):
    dataset_to_validate = InputData.from_csv(data_path)
    predicted_labels = model.predict(dataset_to_validate).predict

    roc_auc_valid = round(roc_auc(y_true=test_data.target,
                                  y_score=predicted_labels,
                                  multi_class='ovo',
                                  average='macro'), 3)
    return roc_auc_valid


if __name__ == '__main__':
    file_path_first = r'./data/example1.xlsx'
    file_path_second = r'./data/example2.xlsx'
    file_path_third = r'./data/example3.xlsx'

    train_file_path, test_file_path = create_multi_clf_examples_from_excel(file_path_first)
    test_data = InputData.from_csv(test_file_path)

    fitted_model = get_model(train_file_path)

    ComposerVisualiser.visualise(fitted_model)

    roc_auc = validate_model_quality(fitted_model, test_file_path)
    print(f'ROC AUC metric is {roc_auc}')

    final_prediction_first = apply_model_to_data(fitted_model, file_path_second)
    print(final_prediction_first['forecast'])

    final_prediction_second = apply_model_to_data(fitted_model, file_path_third)
    print(final_prediction_second['forecast'])



import numpy as np
from sklearn import preprocessing
from sklearn.impute import SimpleImputer

from fedot.core.repository.dataset_types import DataTypesEnum


class PreprocessingStrategy:
    def fit(self, data_to_fit) -> 'PreprocessingStrategy':
        raise NotImplementedError()

    def apply(self, data):
        raise NotImplementedError()


class Scaling(PreprocessingStrategy):
    def __init__(self, with_imputation=True):
        if with_imputation:
            self.default = ImputationStrategy()
        self.with_imputation = with_imputation
        self.scaler = preprocessing.StandardScaler()

    def fit(self, data_to_fit):
        if self.with_imputation:
            self.default.fit(data_to_fit)
            data_to_fit = self.default.apply(data_to_fit)

        data_to_fit = _expand_data(data_to_fit)
        self.scaler.fit(data_to_fit)
        return self

    def apply(self, data):
        if self.with_imputation:
            data = self.default.apply(data)

        data = _expand_data(data)
        resulted = self.scaler.transform(data)
        return resulted


class Normalization(PreprocessingStrategy):
    def __init__(self):
        self.default = ImputationStrategy()

    def fit(self, data_to_fit):
        self.default.fit(data_to_fit)
        return self

    def apply(self, data):
        modified = self.default.apply(data)
        resulted = preprocessing.normalize(modified)

        return resulted


class ImputationStrategy(PreprocessingStrategy):
    def __init__(self):
        self.imputer = SimpleImputer(missing_values=np.nan, strategy='mean')

    def fit(self, data_to_fit):
        self.imputer.fit(data_to_fit)
        return self

    def apply(self, data):
        modified = self.imputer.transform(data)
        return modified


class EmptyStrategy(PreprocessingStrategy):
    def fit(self, data_to_fit):
        return self

    def apply(self, data):
        result = np.asarray(data)
        if len(result.shape) == 1:
            result = np.expand_dims(result, axis=1)
        return result


class TsScalingStrategy(Scaling):
    def __init__(self):
        # the NaN preservation is important for the lagged ts features and forecasted ts
        super().__init__(with_imputation=False)


_preprocessing_for_input_data = {
    DataTypesEnum.ts: EmptyStrategy,
    DataTypesEnum.table: Scaling,
    DataTypesEnum.ts_lagged_table: TsScalingStrategy,
    DataTypesEnum.forecasted_ts: TsScalingStrategy,
}


def preprocessing_func_for_data(data: 'InputData', node: 'Node'):
    preprocessing_func = EmptyStrategy
    if 'without_preprocessing' not in node.model.metadata.tags:
        if node.manual_preprocessing_func:
            preprocessing_func = node.manual_preprocessing_func
        else:
            preprocessing_func = _preprocessing_for_input_data[data.data_type]
    return preprocessing_func


def _expand_data(data):
    if len(data.shape) == 1:
        data = data[:, None]
    return data



import json
import os
import joblib

from uuid import uuid4

from typing import List

from fedot.core.composer.node import PrimaryNode, SecondaryNode, Node, CachedState
from fedot.core.utils import default_fedot_data_dir

DEFAULT_FITTED_MODELS_PATH = os.path.join(default_fedot_data_dir(), 'fitted_models')


class ChainTemplate:
    def __init__(self, chain):
        self.total_chain_types = {}
        self.depth = None
        self.model_templates = []
        self.unique_chain_id = str(uuid4())

        if chain.root_node:
            self._chain_to_template(chain)
        else:
            self.link_to_empty_chain = chain

    def _chain_to_template(self, chain):
        self._extract_chain_structure(chain.root_node, 0, [])
        self.depth = chain.depth

    def _extract_chain_structure(self, node: Node, model_id: int, visited_nodes: List[str]):
        """
        Recursively go through the Chain from 'root_node' to PrimaryNode's,
        creating a ModelTemplate with unique id for each Node. In addition,
        checking whether this Node has been visited yet.
        """
        if node.nodes_from:
            nodes_from = []
            for node_parent in node.nodes_from:
                if node_parent.descriptive_id in visited_nodes:
                    nodes_from.append(visited_nodes.index(node_parent.descriptive_id) + 1)
                else:
                    visited_nodes.append(node_parent.descriptive_id)
                    nodes_from.append(len(visited_nodes))
                    self._extract_chain_structure(node_parent, len(visited_nodes), visited_nodes)
        else:
            nodes_from = []

        model_template = ModelTemplate(node, model_id, nodes_from, self.unique_chain_id)

        self.model_templates.append(model_template)
        self._add_chain_type_to_state(model_template.model_type)

        return model_template

    def _add_chain_type_to_state(self, model_type: str):
        if model_type in self.total_chain_types:
            self.total_chain_types[model_type] += 1
        else:
            self.total_chain_types[model_type] = 1

    def export_to_json(self, path: str):
        path = self._create_unique_chain_id(path)
        absolute_path = os.path.abspath(path)
        if not os.path.exists(absolute_path):
            os.makedirs(absolute_path)

        chain_template_dict = self.convert_to_dict()
        json_data = json.dumps(chain_template_dict)
        with open(os.path.join(absolute_path, f'{self.unique_chain_id}.json'), 'w', encoding='utf-8') as f:
            f.write(json.dumps(json.loads(json_data), indent=4))
            resulted_path = os.path.join(absolute_path, f'{self.unique_chain_id}.json')
            print(f"The chain saved in the path: {resulted_path}")

        return json_data

    def _create_unique_chain_id(self, path_to_save: str):
        name_of_files = path_to_save.split('/')
        last_file = name_of_files[-1].split('.')
        if len(last_file) >= 2:
            if last_file[-1] == 'json':
                if os.path.exists(path_to_save):
                    raise FileExistsError(f"File on path: '{os.path.abspath(path_to_save)}' exists")
                self.unique_chain_id = ''.join(last_file[0:-1])
                return '/'.join(name_of_files[0:-1])
            else:
                raise ValueError(f"Could not save chain in"
                                 f" '{last_file[-1]}' extension, use 'json' format")
        else:
            return path_to_save

    def convert_to_dict(self) -> dict:
        sorted_chain_types = self.total_chain_types
        json_nodes = list(map(lambda model_template: model_template.export_to_json(), self.model_templates))

        json_object = {
            "total_chain_types": sorted_chain_types,
            "depth": self.depth,
            "nodes": json_nodes,
        }

        return json_object

    def import_from_json(self, path: str):
        self._check_for_json_existence(path)

        with open(path) as json_file:
            json_object_chain = json.load(json_file)

        self._extract_models(json_object_chain)
        self.convert_to_chain(self.link_to_empty_chain)
        self.depth = self.link_to_empty_chain.depth
        self.link_to_empty_chain = None

    def _check_for_json_existence(self, path: str):
        absolute_path = os.path.abspath(path)
        path, name_of_file = os.path.split(path)
        name_of_file = name_of_file.split('.')
        if len(name_of_file) >= 2:
            if name_of_file[-1] == 'json':
                if not os.path.isfile(absolute_path):
                    raise FileNotFoundError(f"File on the path: {absolute_path} does not exist.")
                self.unique_chain_id = ''.join(name_of_file[0:-1])
            else:
                raise ValueError(f"Could not load chain in"
                                 f" '{name_of_file[-1]}' extension, use 'json' format")
        else:
            raise FileNotFoundError(f"Write path to 'json' format file")

    def _extract_models(self, chain_json):
        model_objects = chain_json['nodes']

        for model_object in model_objects:
            model_template = ModelTemplate()
            model_template.import_from_json(model_object)
            self._add_chain_type_to_state(model_template.model_type)
            self.model_templates.append(model_template)

    def convert_to_chain(self, chain_to_convert_to: 'Chain'):
        visited_nodes = {}
        root_template = [model_template for model_template in self.model_templates if model_template.model_id == 0][0]
        root_node = _roll_chain_structure(root_template, visited_nodes, self)
        chain_to_convert_to.nodes.clear()
        chain_to_convert_to.add_node(root_node)


def _roll_chain_structure(model_object: 'ModelTemplate', visited_nodes: dict, chain_object: ChainTemplate):
    if model_object.model_id in visited_nodes:
        return visited_nodes[model_object.model_id]
    if model_object.nodes_from:
        node = SecondaryNode(model_object.model_type)
    else:
        node = PrimaryNode(model_object.model_type)

    node.model.params = model_object.params
    nodes_from = [model_template for model_template in chain_object.model_templates
                  if model_template.model_id in model_object.nodes_from]
    node.nodes_from = [_roll_chain_structure(node_from, visited_nodes, chain_object) for node_from
                       in nodes_from]

    if model_object.fitted_model_path:
        path_to_model = os.path.abspath(model_object.fitted_model_path)
        if not os.path.isfile(path_to_model):
            raise FileNotFoundError(f"File on the path: {path_to_model} does not exist.")

        fitted_model = joblib.load(path_to_model)
        node.cache.append(CachedState(preprocessor=model_object.preprocessor,
                                      model=fitted_model))
    visited_nodes[model_object.model_id] = node
    return node


class ModelTemplate:
    def __init__(self, node: Node = None, model_id: int = None,
                 nodes_from: list = None, chain_id: str = None):
        self.model_id = None
        self.model_type = None
        self.model_name = None
        self.custom_params = None
        self.params = None
        self.nodes_from = None
        self.fitted_model_path = None
        self.preprocessor = None

        if node:
            self._model_to_template(node, model_id, nodes_from, chain_id)

    def _model_to_template(self, node: Node, model_id: int, nodes_from: list, chain_id: str):
        self.model_id = model_id
        self.model_type = node.model.model_type
        self.custom_params = node.model.params
        self.params = self._create_full_params(node)
        self.nodes_from = nodes_from

        if _is_node_fitted(node):
            self.model_name = _extract_model_name(node)
            self._extract_fitted_model(node, chain_id)
            self.preprocessor = _extract_preprocessing_strategy(node)

    def _extract_fitted_model(self, node: Node, chain_id: str):
        absolute_path = os.path.abspath(os.path.join(DEFAULT_FITTED_MODELS_PATH, chain_id))
        if not os.path.exists(absolute_path):
            os.makedirs(absolute_path)
        model_name = f'model_{str(self.model_id)}.pkl'
        self.fitted_model_path = os.path.join(DEFAULT_FITTED_MODELS_PATH, chain_id, model_name)
        joblib.dump(node.cache.actual_cached_state.model, self.fitted_model_path)

    def _create_full_params(self, node: Node) -> dict:
        params = {}
        if _is_node_fitted(node):
            params = _extract_model_params(node)
            if isinstance(self.custom_params, dict):
                for key, value in self.custom_params.items():
                    params[key] = value

        return params

    def export_to_json(self) -> dict:

        model_object = {
            "model_id": self.model_id,
            "model_type": self.model_type,
            "model_name": self.model_name,
            "custom_params": self.custom_params,
            "params": self.params,
            "nodes_from": self.nodes_from,
            "trained_model_path": self.fitted_model_path,
        }

        return model_object

    def import_from_json(self, model_object: dict):
        _validate_json_model_template(model_object)

        self.model_id = model_object['model_id']
        self.model_type = model_object['model_type']
        self.params = model_object['params']
        self.nodes_from = model_object['nodes_from']
        if "trained_model_path" in model_object:
            self.fitted_model_path = model_object['trained_model_path']
        if "custom_params" in model_object:
            self.custom_params = model_object['custom_params']
        if "model_name" in model_object:
            self.model_name = model_object['model_name']


def _validate_json_model_template(model_object: dict):
    required_fields = ['model_id', 'model_type', 'params', 'nodes_from']

    for field in required_fields:
        if field not in model_object:
            raise RuntimeError(f"Required field '{field}' is expected, but not found")


def _extract_model_params(node: Node):
    return node.cache.actual_cached_state.model.get_params()


def _extract_model_name(node: Node):
    return node.cache.actual_cached_state.model.__class__.__name__


def _is_node_fitted(node: Node) -> bool:
    return bool(node.cache.actual_cached_state)


def _extract_preprocessing_strategy(node: Node):
    return node.cache.actual_cached_state.preprocessor


def extract_subtree_root(root_model_id: int, chain_template: ChainTemplate):
    root_node = [model_template for model_template in chain_template.model_templates
                 if model_template.model_id == root_model_id][0]
    root_node = _roll_chain_structure(root_node, {}, chain_template)

    return root_node



import operator
from datetime import timedelta
from typing import Callable, Tuple, Union

from numpy.random import choice as nprand_choice, randint
from sklearn.metrics import make_scorer, mean_squared_error, roc_auc_score
from sklearn.metrics import mean_squared_error as mse
from sklearn.model_selection import GridSearchCV, RandomizedSearchCV, cross_val_score
from skopt import BayesSearchCV

from fedot.core.composer.timer import TunerTimer
from fedot.core.log import default_log, Log
from fedot.core.models.data import InputData, train_test_data_setup
from fedot.core.models.tuning.tuner_adapter import HyperoptAdapter
from fedot.core.repository.tasks import TaskTypesEnum

TUNER_ERROR_PREFIX = 'Unsuccessful fit because of'


class Tuner:
    """
    Base class for tuning strategy

    :param trained_model: trained model object
    :param tune_data: data used for hyperparameter searching
    :param dict params_range: search space for hyperparameters
    :param int cross_val_fold_num: number of folds used in cross validation
    :param time_limit: max time available for tuning process
    :param int iterations: max number of iterations
    :param Log log: Log object to record messages
    """
    __tuning_metric_by_type = {
        TaskTypesEnum.classification:
            make_scorer(roc_auc_score, greater_is_better=True, needs_proba=True),
        TaskTypesEnum.regression:
            make_scorer(mean_squared_error, greater_is_better=False),
    }

    def __init__(self, trained_model, tune_data: InputData,
                 params_range: dict,
                 cross_val_fold_num: int,
                 time_limit,
                 iterations: int,
                 log: Log = None):
        self.time_limit: timedelta \
            = time_limit
        self.trained_model = trained_model
        self.tune_data = tune_data
        self.params_range = params_range
        self.cross_val_fold_num = cross_val_fold_num
        self.scorer = self.__tuning_metric_by_type.get(self.tune_data.task.task_type, None)
        self.max_iterations = iterations
        self.default_score, self.default_params = \
            self.get_cross_val_score_and_params(self.trained_model)

        if not log:
            self.log = default_log(__name__)
        else:
            self.log = log

    def tune(self) -> Union[Tuple[dict, object], Tuple[None, None]]:
        """
        Main function to execute hyperparameter search process

        :rtype: Union[Tuple[dict, object], Tuple[None, None]]
        :return: tuple of found hyperparameters and trained model object
        """
        raise NotImplementedError()

    def _is_score_better(self, previous, current):
        __compare = {
            TaskTypesEnum.classification: operator.gt,
            TaskTypesEnum.regression: operator.lt
        }
        comparison = __compare.get(self.tune_data.task.task_type)
        try:
            return comparison(current, previous)
        except Exception as ex:
            self.log.error(f'Score comparison can not be held because {ex}')
            return None, None

    def is_better_than_default(self, score):
        return self._is_score_better(self.default_score, score)

    def get_cross_val_score_and_params(self, model):
        score = cross_val_score(model, self.tune_data.features,
                                self.tune_data.target, scoring=self.scorer,
                                cv=self.cross_val_fold_num).mean()
        params = model.get_params()

        return score, params


class SklearnTuner(Tuner):
    """
    Base tuning strategy used for sklearn models

    :param search_strategy: strategy used for hyperparameter searching (i.e. random, Bayes, etc.)
    """
    def __init__(self, trained_model, tune_data: InputData,
                 params_range: dict,
                 cross_val_fold_num: int,
                 time_limit, iterations):
        super().__init__(trained_model=trained_model,
                         tune_data=tune_data,
                         params_range=params_range,
                         cross_val_fold_num=cross_val_fold_num,
                         time_limit=time_limit,
                         iterations=iterations)
        self.search_strategy = None

    def tune(self) -> Union[Tuple[dict, object], Tuple[None, None]]:
        raise NotImplementedError()

    def _sklearn_tune(self, tune_data: InputData):
        try:
            search = self.search_strategy.fit(tune_data.features, tune_data.target.ravel())
            new_score, _ = self.get_cross_val_score_and_params(search.best_estimator_)
            if self.is_better_than_default(new_score):
                return search.best_params_, search.best_estimator_
            else:
                return self.default_params, self.trained_model
        except Exception as ex:
            self.log.error(f'{TUNER_ERROR_PREFIX} {ex}')
            return None, None


class SklearnRandomTuner(SklearnTuner):
    """
    Sklearn tuning strategy using RandomSearchCV
    """
    def tune(self) -> Union[Tuple[dict, object], Tuple[None, None]]:
        self.search_strategy = RandomizedSearchCV(estimator=self.trained_model,
                                                  param_distributions=self.params_range,
                                                  n_iter=self.max_iterations,
                                                  cv=self.cross_val_fold_num,
                                                  scoring=self.scorer)
        return self._sklearn_tune(tune_data=self.tune_data)


class SklearnGridSearchTuner(SklearnTuner):
    """
    Sklearn tuning strategy using GridSearchCV
    """
    def tune(self) -> Union[Tuple[dict, object], Tuple[None, None]]:
        self.search_strategy = GridSearchCV(estimator=self.trained_model,
                                            param_grid=self.params_range,
                                            cv=self.cross_val_fold_num,
                                            scoring=self.scorer)
        return self._sklearn_tune(self.tune_data)


class SklearnBayesSearchCV(SklearnTuner):
    """
    Sklearn tuning strategy using BayesSearchCV
    """
    def tune(self) -> Union[Tuple[dict, object], Tuple[None, None]]:
        self.search_strategy = BayesSearchCV(estimator=self.trained_model,
                                             search_spaces=self.params_range,
                                             n_iter=self.max_iterations,
                                             cv=self.cross_val_fold_num,
                                             scoring=self.scorer)
        return self._sklearn_tune(self.tune_data)


class SklearnCustomRandomTuner(Tuner):
    """
    Sklearn tuning strategy using customized version of RandomSearch with cross validation
    """
    def tune(self) -> Union[Tuple[dict, object], Tuple[None, None]]:
        try:
            with TunerTimer() as timer:
                best_model = self.trained_model
                best_score, best_params = self.get_cross_val_score_and_params(best_model)
                for iteration in range(self.max_iterations):
                    params = {k: nprand_choice(v) for k, v in self.params_range.items()}
                    self.trained_model.set_params(**params)
                    score, _ = self.get_cross_val_score_and_params(self.trained_model)
                    if self._is_score_better(previous=best_score, current=score):
                        best_params = params
                        best_model = self.trained_model
                        best_score = score

                    if timer.is_time_limit_reached(self.time_limit):
                        break
                return best_params, best_model
        except Exception as ex:
            self.log.error(f'{TUNER_ERROR_PREFIX} {ex}')
            return None, None


class ForecastingCustomRandomTuner:
    """
    Tuning strategy used for forecasting models
    """
    def __init__(self, **kwargs):
        if 'log' not in kwargs:
            self.logger = default_log(__name__)
        else:
            self.logger = kwargs['log']

    # TODO discuss
    def tune(self,
             fit: Callable,
             predict: Callable,
             tune_data: InputData, params_range: dict,
             default_params: dict, iterations: int) -> dict:
        """

        :param Callable fit: function used to fit the model
        :param Callable predict: function used to predict the model
        :param InputData tune_data: data used in hyperparameter searching
        :param dict params_range: search space for hyperparameters
        :param sict default_params: default values of model hyperparameters
        :param int iterations: max number of iterations
        :return: best parameters found via tuning
        :rtype: dict
        """
        tune_train_data, tune_test_data = train_test_data_setup(tune_data, 0.5)

        trained_model_default = fit(tune_test_data, default_params)
        prediction_default = predict(trained_model_default, tune_test_data)
        best_quality_metric = _regression_prediction_quality(prediction=prediction_default,
                                                             real=tune_test_data.target)
        best_params = default_params

        for _ in range(iterations):
            random_params = get_random_params(params_range)
            try:
                trained_model_candidate = fit(tune_train_data, random_params)
                prediction_candidate = predict(trained_model_candidate,
                                               tune_test_data)
                quality_metric = _regression_prediction_quality(prediction=prediction_candidate,
                                                                real=tune_test_data.target)
                if quality_metric < best_quality_metric:
                    best_params = random_params
            except Exception as ex:
                self.logger.error(f'{TUNER_ERROR_PREFIX} {ex}')
        return best_params


def get_random_params(params_range):
    candidate_params = {}
    for param in params_range:
        param_range = params_range[param]
        param_range_left, param_range_right = param_range[0], param_range[1]
        if isinstance(param_range_left, tuple):
            # set-based params with constant length
            candidate_param = get_constant_length_range(param_range_left, param_range_right)
        elif isinstance(param_range_left, list):
            # set-based params with varied length
            candidate_param = get_varied_length_range(param_range_left, param_range_right)
        else:
            raise ValueError(f'Un-supported params range type {type(param_range_left)}')
        candidate_params[param] = candidate_param
    return candidate_params


def get_constant_length_range(left_range, right_range):
    candidate_param = []
    for sub_param_ind in range(len(left_range)):
        new_sub_param = randint(left_range[sub_param_ind],
                                right_range[sub_param_ind] + 1)
        candidate_param.append(new_sub_param)
    return tuple(candidate_param)


def get_varied_length_range(left_range, right_range):
    candidate_param = []
    subparams_num = randint(1, len(right_range))
    for sub_param_ind in range(subparams_num):
        new_sub_param = randint(left_range[sub_param_ind],
                                right_range[sub_param_ind] + 1)
        candidate_param.append(new_sub_param)
    return candidate_param


def _regression_prediction_quality(prediction, real):
    return mse(y_true=real, y_pred=prediction, squared=False)


class TPETuner(Tuner):
    """
    Tuning strategy using Tree Parzen Estimator from hyperopt library
    """
    def tune(self) -> Union[Tuple[dict, object], Tuple[None, None]]:
        try:
            adapter = HyperoptAdapter(self)
            best_params, best_model = adapter.tune(iterations=self.max_iterations,
                                                   timeout_sec=self.time_limit.seconds)
            new_score, _ = self.get_cross_val_score_and_params(best_model)

            if self.is_better_than_default(new_score):
                return best_params, best_model
            else:
                return self.default_params, self.trained_model
        except Exception as ex:
            self.log.error(f'{TUNER_ERROR_PREFIX} {ex}')
            return None, None



from copy import copy

import matplotlib.pyplot as plt
import numpy as np

from fedot.core.composer.node import PrimaryNode, SecondaryNode
from fedot.core.composer.ts_chain import TsForecastingChain
from fedot.core.models.data import InputData, train_test_data_setup
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.tasks import Task, TaskTypesEnum, TsForecastingParams


def get_synthetic_ts_data_custom(n_steps=6000, forecast_length=2, max_window_size=2,
                                 with_exog: bool = True) -> InputData:
    task = Task(TaskTypesEnum.ts_forecasting,
                TsForecastingParams(forecast_length=forecast_length,
                                    max_window_size=max_window_size,
                                    return_all_steps=False,
                                    make_future_prediction=True))

    exog_features = np.asarray([10.0, 11.0, 12.0, 13.0, 14.0, 15.0, 16.0, 17.0])
    if not with_exog:
        # move target to features
        exog_features = None

    input_data = InputData(idx=np.arange(0, n_steps),
                           features=exog_features,
                           target=np.asarray([0.0, 1.0, 2.0, 3.0, 4.0, 5.0]),
                           task=task,
                           data_type=DataTypesEnum.ts)
    return input_data


def get_synthetic_ts_data_period(n_steps=6000, forecast_length=1, max_window_size=50,
                                 with_exog: bool = True) -> InputData:
    x1 = np.arange(0, n_steps) / 10
    x2 = np.arange(0, n_steps) + 1

    x1_exog = np.arange(0, n_steps + forecast_length) / 10
    x2_exog = np.arange(0, n_steps + forecast_length) + 1

    simulated_data = x1 * 0.005 - x2 * 0.001
    periodicity = np.sin(x1 * 0.4)
    random = np.random.normal(0, 0.1, n_steps)
    simulated_data = simulated_data + periodicity + random

    task = Task(TaskTypesEnum.ts_forecasting,
                TsForecastingParams(forecast_length=forecast_length,
                                    max_window_size=max_window_size,
                                    return_all_steps=False,
                                    make_future_prediction=True))

    exog_features = np.asarray([x1_exog, x2_exog]).T
    if not with_exog:
        # move target to features
        exog_features = None
    input_data = InputData(idx=np.arange(0, n_steps),
                           features=exog_features,
                           target=simulated_data,
                           task=task,
                           data_type=DataTypesEnum.ts)
    return input_data


def plot_prediction(prediction, data: InputData, desc) -> (float, float):
    plt.plot(data.target, label='target')
    plt.plot(prediction, label='predict')
    plt.legend()
    # plt.ylim(min(data.target), max(data.target))
    plt.title(desc)
    plt.show()


def run_forecasting(chain: TsForecastingChain, data: InputData, is_visualise: bool, desc: str):
    train_data, test_data = train_test_data_setup(data, shuffle_flag=False, split_ratio=0.9)
    data.task.task_params.make_future_prediction = False
    chain.fit_from_scratch(train_data)

    test_data_for_pred = copy(test_data)
    # to avoid data leak
    test_data_for_pred.target = None
    data.task.task_params.make_future_prediction = True

    full_prediction = chain.forecast(initial_data=train_data, supplementary_data=test_data_for_pred).predict
    if is_visualise:
        plot_prediction(full_prediction, test_data, desc)


def run_onestep_linear_example(n_steps=1000, is_visualise: bool = True):
    window_size = 16

    dataset = get_synthetic_ts_data_period(n_steps=n_steps,
                                           forecast_length=1,
                                           max_window_size=window_size,
                                           with_exog=True)
    # regression forecasting
    chain = TsForecastingChain(PrimaryNode('linear'))

    # one step regression
    run_forecasting(chain=chain, data=dataset, is_visualise=is_visualise,
                    desc=f'Linear model, {dataset.task.task_params.forecast_length} step prediction with exog')

    dataset = get_synthetic_ts_data_period(n_steps=n_steps,
                                           forecast_length=1,
                                           max_window_size=window_size,
                                           with_exog=False)

    run_forecasting(chain=chain, data=dataset, is_visualise=is_visualise,
                    desc=f'Linear model, {dataset.task.task_params.forecast_length} step prediction without exog')


def run_multistep_linear_example(n_steps=6000, is_visualise: bool = True):
    chain = TsForecastingChain(PrimaryNode('ridge'))

    dataset = get_synthetic_ts_data_period(n_steps=n_steps,
                                           forecast_length=64,
                                           max_window_size=64,
                                           with_exog=True)
    # multi step regression
    run_forecasting(chain=chain, data=dataset, is_visualise=is_visualise,
                    desc=f'Linear model, {dataset.task.task_params.forecast_length} step prediction with exog')

    dataset = get_synthetic_ts_data_period(n_steps=n_steps,
                                           forecast_length=64,
                                           max_window_size=64,
                                           with_exog=False)
    run_forecasting(chain=chain, data=dataset, is_visualise=is_visualise,
                    desc=f'Linear model, {dataset.task.task_params.forecast_length} step prediction without exog')


def run_multistep_custom_example(n_steps=6, is_visualise: bool = True):
    chain = TsForecastingChain(PrimaryNode('ridge'))

    dataset = get_synthetic_ts_data_custom(n_steps=n_steps,
                                           forecast_length=2,
                                           max_window_size=2,
                                           with_exog=False)
    # multi step regression
    run_forecasting(chain=chain, data=dataset, is_visualise=is_visualise,
                    desc=f'Linear model, {dataset.task.task_params.forecast_length} step prediction without exog')

    dataset = get_synthetic_ts_data_custom(n_steps=n_steps,
                                           forecast_length=2,
                                           max_window_size=2,
                                           with_exog=True)
    run_forecasting(chain=chain, data=dataset, is_visualise=is_visualise,
                    desc=f'Linear model, {dataset.task.task_params.forecast_length} step prediction with exog')


def run_multistep_multiscale_example(n_steps=10000, is_visualise: bool = True):
    dataset = get_synthetic_ts_data_period(n_steps=n_steps,
                                           forecast_length=64,
                                           max_window_size=512,
                                           with_exog=False)

    # composite forecasting with decomposition
    node_first = PrimaryNode('trend_data_model')
    node_second = PrimaryNode('residual_data_model')
    node_trend_model = SecondaryNode('ridge', nodes_from=[node_first])
    node_residual_model = SecondaryNode('linear', nodes_from=[node_second])

    node_final = SecondaryNode('linear', nodes_from=[node_trend_model, node_residual_model])

    chain = TsForecastingChain(node_final)

    run_forecasting(chain=chain, data=dataset,
                    is_visualise=is_visualise,
                    desc=f'Multiscale model, {dataset.task.task_params.forecast_length} step prediction withot exog')


def run_multistep_composite_example(n_steps=20000, is_visualise: bool = True):
    # composite forecasting with ensemble
    node_first = PrimaryNode('linear')
    node_second = PrimaryNode('ridge')
    node_final = SecondaryNode('linear', nodes_from=[node_first, node_second])

    chain = TsForecastingChain(node_final)

    dataset = get_synthetic_ts_data_period(n_steps=n_steps,
                                           forecast_length=64,
                                           max_window_size=512,
                                           with_exog=False)

    run_forecasting(chain=chain, data=dataset,
                    is_visualise=is_visualise,
                    desc=f'Composite model, {dataset.task.task_params.forecast_length} step prediction without exog')

    dataset = get_synthetic_ts_data_period(n_steps=n_steps,
                                           forecast_length=64,
                                           max_window_size=64,
                                           with_exog=True)

    run_forecasting(chain=chain, data=dataset,
                    is_visualise=is_visualise,
                    desc=f'Composite model, {dataset.task.task_params.forecast_length} step prediction with exog')


def run_multistep_lstm_example(n_steps=6000, is_visualise: bool = True):
    # lstm forecasting
    dataset = get_synthetic_ts_data_period(n_steps=n_steps,
                                           forecast_length=64,
                                           max_window_size=64,
                                           with_exog=False)

    chain = TsForecastingChain(PrimaryNode('lstm'))
    run_forecasting(chain=chain, data=dataset,
                    is_visualise=is_visualise,
                    desc=f'LSTM model, {dataset.task.task_params.forecast_length} step prediction with exog')

    return True


if __name__ == '__main__':
    print('Onestep linear')
    run_onestep_linear_example()
    print('Multistep linear')
    run_multistep_linear_example()
    print('Multistep multiscale_example')
    run_multistep_multiscale_example()
    print('Multistep composite')
    run_multistep_composite_example()
    print('Multistep LSTM')
    run_multistep_lstm_example()



from typing import Callable

from fedot.core.composer.metrics import F1Metric, MaeMetric, RmseMetric, \
    RocAucMetric, StructuralComplexityMetric
from fedot.core.utils import ComparableEnum as Enum


class MetricsEnum(Enum):
    pass


class QualityMetricsEnum(MetricsEnum):
    pass


class ComplexityMetricsEnum(MetricsEnum):
    node_num = 'node_number'
    structural = 'structural'


class ClassificationMetricsEnum(QualityMetricsEnum):
    ROCAUC = 'roc_auc'
    ROCAUC_penalty = 'roc_auc_pen'
    precision = 'precision'
    f1 = 'f1'


class RegressionMetricsEnum(QualityMetricsEnum):
    RMSE = 'rmse'
    RMSE_penalty = 'roc_auc_pen'
    MAE = 'mae'


class MetricsRepository:
    __metrics_implementations = {
        ClassificationMetricsEnum.ROCAUC: RocAucMetric.get_value,
        ClassificationMetricsEnum.ROCAUC_penalty: RocAucMetric.get_value_with_penalty,
        RegressionMetricsEnum.MAE: MaeMetric.get_value,
        RegressionMetricsEnum.RMSE: RmseMetric.get_value,
        RegressionMetricsEnum.RMSE_penalty: RmseMetric.get_value_with_penalty,
        ClassificationMetricsEnum.f1: F1Metric.get_value,
        ComplexityMetricsEnum.structural: StructuralComplexityMetric.get_value
    }

    def metric_by_id(self, metric_id: MetricsEnum) -> Callable:
        return self.__metrics_implementations[metric_id]



from copy import deepcopy
from random import choice, randint
from typing import (Any, List, Tuple)

from fedot.core.composer.constraint import constraint_function


def node_height(chain: Any, node: Any) -> int:
    def recursive_child_height(parent_node: Any) -> int:
        node_child = chain.node_childs(parent_node)
        if node_child:
            height = recursive_child_height(node_child[0]) + 1
            return height
        else:
            return 0

    height = recursive_child_height(node)
    return height


def node_depth(node: Any) -> int:
    if not node.nodes_from:
        return 0
    else:
        return 1 + max([node_depth(next_node) for next_node in node.nodes_from])


def nodes_from_height(chain: Any, selected_height: int) -> List[Any]:
    def get_nodes(node: Any, current_height):
        nodes = []
        if current_height == selected_height:
            nodes.append(node)
        else:
            if node.nodes_from:
                for child in node.nodes_from:
                    nodes += get_nodes(child, current_height + 1)
        return nodes

    nodes = get_nodes(chain.root_node, current_height=0)
    return nodes


def random_chain(chain_generation_params, requirements, max_depth=None) -> Any:
    secondary_node_func = chain_generation_params.secondary_node_func
    primary_node_func = chain_generation_params.primary_node_func
    chain_class = chain_generation_params.chain_class
    max_depth = max_depth if max_depth else requirements.max_depth

    def chain_growth(chain: Any, node_parent: Any):
        offspring_size = randint(requirements.min_arity, requirements.max_arity)
        for offspring_node in range(offspring_size):
            height = node_height(chain, node_parent)
            is_max_depth_exceeded = height >= max_depth - 1
            is_primary_node_selected = height < max_depth - 1 and randint(0, 1)
            if is_max_depth_exceeded or is_primary_node_selected:
                primary_node = primary_node_func(model_type=choice(requirements.primary))
                node_parent.nodes_from.append(primary_node)
                chain.add_node(primary_node)
            else:
                secondary_node = secondary_node_func(model_type=choice(requirements.secondary))
                chain.add_node(secondary_node)
                node_parent.nodes_from.append(secondary_node)
                chain_growth(chain, secondary_node)

    is_correct_chain = False
    while not is_correct_chain:
        chain = chain_class()
        chain_root = secondary_node_func(model_type=choice(requirements.secondary))
        chain.add_node(chain_root)
        chain_growth(chain, chain_root)
        is_correct_chain = constraint_function(chain)
    return chain


def equivalent_subtree(chain_first: Any, chain_second: Any) -> List[Tuple[Any, Any]]:
    """Finds the similar subtree in two given trees"""

    def structural_equivalent_nodes(node_first, node_second):
        nodes = []
        is_same_type = type(node_first) == type(node_second)
        node_first_childs = node_first.nodes_from
        node_second_childs = node_second.nodes_from
        if is_same_type and ((not node_first.nodes_from) or len(node_first_childs) == len(node_second_childs)):
            nodes.append((node_first, node_second))
            if node_first.nodes_from:
                for node1_child, node2_child in zip(node_first.nodes_from, node_second.nodes_from):
                    nodes_set = structural_equivalent_nodes(node1_child, node2_child)
                    if nodes_set:
                        nodes += nodes_set
        return nodes

    pairs_set = structural_equivalent_nodes(chain_first.root_node, chain_second.root_node)
    assert isinstance(pairs_set, list)
    return pairs_set


def replace_subtrees(chain_first: Any, chain_second: Any, node_from_first: Any, node_from_second: Any,
                     layer_in_first: int, layer_in_second: int, max_depth: int):
    node_from_chain_first_copy = deepcopy(node_from_first)

    summary_depth = layer_in_first + node_depth(node_from_second)
    if summary_depth <= max_depth and summary_depth != 0:
        chain_first.replace_node_with_parents(node_from_first, node_from_second)

    summary_depth = layer_in_second + node_depth(node_from_first)
    if summary_depth <= max_depth and summary_depth != 0:
        chain_second.replace_node_with_parents(node_from_second, node_from_chain_first_copy)


def num_of_parents_in_crossover(num_of_final_inds: int) -> int:
    return num_of_final_inds if not num_of_final_inds % 2 else num_of_final_inds + 1



import math
from random import choice, randint
from typing import (Any, List)

from fedot.core.utils import ComparableEnum as Enum


class SelectionTypesEnum(Enum):
    tournament = 'tournament'


def selection(types: List[SelectionTypesEnum], population: List[Any], pop_size: int) -> List[Any]:
    selection_by_type = {
        SelectionTypesEnum.tournament: tournament_selection,
    }

    type = choice(types)
    if type in selection_by_type.keys():
        return selection_by_type[type](population, pop_size)
    else:
        raise ValueError(f'Required selection not found: {type}')


def individuals_selection(types: List[SelectionTypesEnum], individuals: List[Any], pop_size: int) -> List[Any]:
    if pop_size == len(individuals):
        chosen = individuals
    else:
        chosen = []
        remaining_individuals = individuals
        for _ in range(pop_size):
            individual = selection(types, remaining_individuals, pop_size=1)[0]
            chosen.append(individual)
            remaining_individuals.remove(individual)
    return chosen


def random_selection(individuals: List[Any], pop_size: int) -> List[int]:
    return [individuals[randint(0, len(individuals) - 1)] for _ in range(pop_size)]


def tournament_selection(individuals: List[Any], pop_size: int, fraction: float = 0.1) -> List[Any]:
    group_size = math.ceil(len(individuals) * fraction)
    min_group_size = 2 if len(individuals) > 1 else 1
    group_size = max(group_size, min_group_size)
    chosen = []
    for _ in range(pop_size):
        group = random_selection(individuals, group_size)
        best = min(group, key=lambda ind: ind.fitness)
        chosen.append(best)
    return chosen



import warnings
from dataclasses import dataclass
from typing import List, Optional, Tuple

import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split

from fedot.core.algorithms.time_series.lagged_features import prepare_lagged_ts_for_prediction
from fedot.core.models.preprocessing import ImputationStrategy
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.tasks import Task, TaskTypesEnum


@dataclass
class Data:
    """
    Base Data type class
    """
    idx: np.array
    features: np.array
    task: Task
    data_type: DataTypesEnum

    @staticmethod
    def from_csv(file_path=None,
                 delimiter=',',
                 task: Task = Task(TaskTypesEnum.classification),
                 data_type: DataTypesEnum = DataTypesEnum.table,
                 columns_to_drop: Optional[List] = None,
                 target_column: Optional[str] = ''):
        """
        :param file_path: the path to the CSV with data
        :param columns_to_drop: the names of columns that should be dropped
        :param delimiter: the delimiter to separate the columns
        :param task: the task that should be solved with data
        :param data_type: the type of data interpretation
        :param target_column: name of target column (last column if empty and no target if None)
        :return:
        """

        data_frame = pd.read_csv(file_path, sep=delimiter)
        if columns_to_drop:
            data_frame = data_frame.drop(columns_to_drop, axis=1)
        data_frame = _convert_dtypes(data_frame=data_frame)
        data_array = np.array(data_frame).T
        idx = data_array[0]

        if target_column == '':
            target_column = data_frame.columns[-1]

        if target_column:
            target = np.array(data_frame[target_column]).astype(np.float)
            pos = list(data_frame.keys()).index(target_column)
            features = np.delete(data_array.T, [0, pos], axis=1)
        else:
            # no target in data
            features = data_array[1:].T
            target = None

        features = ImputationStrategy().fit(features).apply(features)

        return InputData(idx=idx, features=features, target=target, task=task, data_type=data_type)


@dataclass
class InputData(Data):
    """
    Data class for input data for the nodes
    """
    target: Optional[np.array] = None

    @property
    def num_classes(self) -> Optional[int]:
        if self.task.task_type == TaskTypesEnum.classification:
            return len(np.unique(self.target))
        else:
            return None

    @staticmethod
    def from_predictions(outputs: List['OutputData'], target: np.array):
        if len(set([output.task.task_type for output in outputs])) > 1:
            raise ValueError('Inconsistent task types')

        task = outputs[0].task
        data_type = outputs[0].data_type
        idx = outputs[0].idx

        dataset_merging_funcs = {
            DataTypesEnum.forecasted_ts: _combine_datasets_ts,
            DataTypesEnum.ts: _combine_datasets_ts,
            DataTypesEnum.table: _combine_datasets_table
        }
        dataset_merging_funcs.setdefault(data_type, _combine_datasets_common)

        features = dataset_merging_funcs[data_type](outputs)

        return InputData(idx=idx, features=features, target=target, task=task,
                         data_type=data_type)

    def subset(self, start: int, end: int):
        if not (0 <= start <= end <= len(self.idx)):
            raise ValueError('Incorrect boundaries for subset')
        new_features = None
        if self.features is not None:
            new_features = self.features[start:end + 1]
        return InputData(idx=self.idx[start:end + 1], features=new_features,
                         target=self.target[start:end + 1], task=self.task, data_type=self.data_type)

    def prepare_for_modelling(self, is_for_fit: bool = False):
        prepared_data = self
        if (self.data_type == DataTypesEnum.ts_lagged_table or
                self.data_type == DataTypesEnum.forecasted_ts):
            prepared_data = prepare_lagged_ts_for_prediction(self, is_for_fit)
        elif self.data_type in [DataTypesEnum.table, DataTypesEnum.forecasted_ts]:
            # TODO implement NaN filling here
            pass

        return prepared_data


@dataclass
class OutputData(Data):
    """
    Data type for data predicted in the node
    """
    predict: np.array = None


def split_train_test(data, split_ratio=0.8, with_shuffle=False):
    assert 0. <= split_ratio <= 1.
    if with_shuffle:
        data_train, data_test = train_test_split(data, test_size=1. - split_ratio, random_state=42)
    else:
        split_point = int(len(data) * split_ratio)
        data_train, data_test = data[:split_point], data[split_point:]
    return data_train, data_test


def _convert_dtypes(data_frame: pd.DataFrame):
    objects: pd.DataFrame = data_frame.select_dtypes('object')
    for column_name in objects:
        warnings.warn(f'Automatic factorization for the column {column_name} with type "object" is applied.')
        encoded = pd.factorize(data_frame[column_name])[0]
        data_frame[column_name] = encoded
    data_frame = data_frame.fillna(0)
    return data_frame


def train_test_data_setup(data: InputData, split_ratio=0.8, shuffle_flag=False) -> Tuple[InputData, InputData]:
    if data.features is not None:
        train_data_x, test_data_x = split_train_test(data.features, split_ratio, with_shuffle=shuffle_flag)
    else:
        train_data_x, test_data_x = None, None

    train_data_y, test_data_y = split_train_test(data.target, split_ratio, with_shuffle=shuffle_flag)
    train_idx, test_idx = split_train_test(data.idx, split_ratio)
    train_data = InputData(features=train_data_x, target=train_data_y,
                           idx=train_idx, task=data.task, data_type=data.data_type)
    test_data = InputData(features=test_data_x, target=test_data_y, idx=test_idx, task=data.task,
                          data_type=data.data_type)
    return train_data, test_data


def _combine_datasets_ts(outputs: List[OutputData]):
    features_list = list()

    expected_len = max([len(output.predict) for output in outputs])

    for elem in outputs:
        predict = elem.predict
        if len(elem.predict) != expected_len:
            raise ValueError(f'Non-equal prediction length: {len(elem.predict)} and {expected_len}')
        features_list.append(predict)

    if len(features_list) > 1:
        features = np.column_stack(features_list)
    else:
        features = features_list[0]

    return features


def _combine_datasets_table(outputs: List[OutputData]):
    features = list()
    expected_len = len(outputs[0].predict)

    for elem in outputs:
        if len(elem.predict) != expected_len:
            raise ValueError(f'Non-equal prediction length: {len(elem.predict)} and {expected_len}')
        if len(elem.predict.shape) == 1:
            features.append(elem.predict)
        else:
            # if the model prediction is multivariate
            number_of_variables_in_prediction = elem.predict.shape[1]
            for i in range(number_of_variables_in_prediction):
                features.append(elem.predict[:, i])

    features = np.array(features).T

    return features


def _combine_datasets_common(outputs: List[OutputData]):
    features = list()

    for elem in outputs:
        if len(elem.predict) != len(outputs[0].predict):
            raise NotImplementedError(f'Non-equal prediction length: '
                                      f'{len(elem.predict)} and {len(outputs[0].predict)}')
        if len(elem.predict.shape) == 1:
            features.append(elem.predict)
        else:
            # if the model prediction is multivariate
            number_of_variables_in_prediction = elem.predict.shape[1]
            for i in range(number_of_variables_in_prediction):
                features.append(elem.predict[:, i])
    return features



from datetime import timedelta

from sklearn.metrics import roc_auc_score as roc_auc

from cases.data.data_utils import get_scoring_case_data_paths
from fedot.core.composer.chain import Chain
from fedot.core.composer.node import PrimaryNode, SecondaryNode
from fedot.core.models.data import InputData


def run_chain_from_automl(train_file_path: str, test_file_path: str,
                          max_run_time: timedelta = timedelta(minutes=10)):
    train_data = InputData.from_csv(train_file_path)
    test_data = InputData.from_csv(test_file_path)

    testing_target = test_data.target

    chain = Chain()
    node_tpot = PrimaryNode('tpot')

    node_tpot.model.params = {'max_run_time_sec': max_run_time.seconds}

    node_lda = PrimaryNode('lda')
    node_rf = SecondaryNode('rf')

    node_rf.nodes_from = [node_tpot, node_lda]

    chain.add_node(node_rf)

    chain.fit(train_data)
    results = chain.predict(test_data)

    roc_auc_value = roc_auc(y_true=testing_target,
                            y_score=results.predict)
    print(roc_auc_value)

    return roc_auc_value


if __name__ == '__main__':
    train_file_path, test_file_path = get_scoring_case_data_paths()
    run_chain_from_automl(train_file_path, test_file_path)



from fedot.core.chain_validation import validate
from fedot.core.composer.chain import Chain


def constraint_function(chain: Chain):
    try:
        validate(chain)
        return True
    except ValueError:
        return False



import os

import pandas as pd

from fedot.core.repository.model_types_repository import ModelTypesRepository
from fedot.core.repository.tasks import TaskTypesEnum
from fedot.core.utils import ensure_directory_exists, get_split_data_paths, \
    project_root, save_file_to_csv, split_data


def create_multi_clf_examples_from_excel(file_path: str, return_df: bool = False):
    df = pd.read_excel(file_path)
    train, test = split_data(df)
    file_dir_name = file_path.replace('.', '/').split('/')[-2]
    file_csv_name = f'{file_dir_name}.csv'
    directory_names = ['examples', 'data', file_dir_name]
    ensure_directory_exists(directory_names)
    if return_df:
        path = os.path.join(directory_names[0], directory_names[1], directory_names[2], file_csv_name)
        full_file_path = os.path.join(str(project_root()), path)
        save_file_to_csv(df, full_file_path)
        return df, full_file_path
    else:
        full_train_file_path, full_test_file_path = get_split_data_paths(directory_names)
        save_file_to_csv(train, full_train_file_path)
        save_file_to_csv(train, full_test_file_path)
        return full_train_file_path, full_test_file_path


def print_models_info(repository: ModelTypesRepository, task=TaskTypesEnum.classification):
    for model in repository.models:
        print(f'{model.id}, {model.current_strategy(task)}, '
              f'{model.current_strategy(task)(model.id).implementation_info}')


if __name__ == '__main__':
    print_models_info(ModelTypesRepository())



from copy import copy
from functools import partial
from random import randint
from typing import (Any, Callable, List, Optional)

from numpy import random

from fedot.core.composer.chain import Chain, Node
from fedot.core.composer.composer import ComposerRequirements
from fedot.core.composer.node import PrimaryNode, SecondaryNode
from fedot.core.models.data import InputData


class RandomSearchComposer:
    def __init__(self, iter_num: int = 10):
        self.__iter_num = iter_num

    def compose_chain(self, data: InputData,
                      initial_chain: Optional[Chain],
                      composer_requirements: ComposerRequirements,
                      metrics: Optional[Callable]) -> Chain:
        # TODO: fix this later?
        train_data = data
        test_data = data
        metric_function_for_nodes = partial(metric_for_nodes,
                                            metric_function=metrics,
                                            train_data=train_data,
                                            test_data=test_data)

        optimiser = RandomSearchOptimiser(self.__iter_num, PrimaryNode, SecondaryNode)
        best_nodes_set, history = optimiser.optimise(metric_function_for_nodes,
                                                     composer_requirements.primary,
                                                     composer_requirements.secondary)

        best_chain = Chain()
        [best_chain.add_node(nodes) for nodes in best_nodes_set]

        return best_chain


def nodes_to_chain(nodes: List[Node]) -> Chain:
    chain = Chain()
    [chain.add_node(nodes) for nodes in nodes]
    return chain


def metric_for_nodes(metric_function, nodes: List[Node], train_data: InputData, test_data: InputData) -> float:
    chain = nodes_to_chain(nodes)
    chain.fit(input_data=train_data)
    return metric_function(chain, test_data)


class RandomSearchOptimiser:
    def __init__(self, iter_num: int, primary_node_func: Callable, secondary_node_func: Callable):
        self.__iter_num = iter_num
        self.__primary_node_func = primary_node_func
        self.__secondary_node_func = secondary_node_func

    def optimise(self, metric_function_for_nodes,
                 primary_candidates: List[Any],
                 secondary_candidates: List[Any]):
        best_metric_value = 1000
        best_set = []
        history = []
        for i in range(self.__iter_num):
            print(f'Iter {i}')
            new_nodeset = self._random_nodeset(primary_candidates, secondary_candidates)
            new_metric_value = round(metric_function_for_nodes(nodes=new_nodeset), 3)
            history.append((new_nodeset, new_metric_value))

            print(f'Try {new_metric_value} with length {len(new_nodeset)}')
            if new_metric_value < best_metric_value:
                best_metric_value = new_metric_value
                best_set = new_nodeset
                print(f'Better chain found: metric {best_metric_value}')

        return best_set, history

    def _random_nodeset(self, primary_requirements: List[Any], secondary_requirements: List[Any]) -> List[Node]:
        new_set = []

        # random primary nodes
        num_of_primary = randint(1, len(primary_requirements))
        random_first_models = random.choice(primary_requirements, num_of_primary, replace=False)
        [new_set.append(self.__primary_node_func(model)) for model in random_first_models]

        # random final node
        if len(new_set) > 1:
            random_final_model = random.choice(secondary_requirements, replace=False)
            parent_nodes = copy(new_set)
            new_set.append(self.__secondary_node_func(random_final_model, parent_nodes))

        return new_set



from copy import copy, deepcopy
from datetime import timedelta
from typing import List, Optional, Union
from uuid import uuid4

import networkx as nx

from fedot.core.composer.node import (FittedModelCache, Node, PrimaryNode, SecondaryNode, SharedCache)
from fedot.core.log import default_log, Log
from fedot.core.models.data import InputData
from fedot.core.repository.tasks import TaskTypesEnum
from fedot.utilities.synthetic.chain_template_new import ChainTemplate

ERROR_PREFIX = 'Invalid chain configuration:'


class Chain:
    """
    Base class used for composite model structure definition

    :param nodes: Node object(s)
    :param log: Log object to record messages

    .. note::
        fitted_on_data stores the data which were used in last chain fitting (equals None if chain hasn't been
        fitted yet)
    """

    def __init__(self, nodes: Optional[Union[Node, List[Node]]] = None,
                 log: Log = None):
        self.nodes = []
        self.log = log
        self.template = None

        if not log:
            self.log = default_log(__name__)
        else:
            self.log = log

        if nodes:
            if isinstance(nodes, list):
                for node in nodes:
                    self.add_node(node)
            else:
                self.add_node(nodes)
        self.fitted_on_data = None

    def fit_from_scratch(self, input_data: InputData, verbose=False):
        """
        Method used for training the chain without using cached information

        :param input_data: data used for model training
        :param verbose: flag used for status printing to console, default False
        """
        # Clean all cache and fit all models
        self.log.info('Fit chain from scratch')
        self.fit(input_data, use_cache=False, verbose=verbose)

    def cache_status_if_new_data(self, new_input_data: InputData, cache_status: bool):
        if self.fitted_on_data is not None and self.fitted_on_data is not new_input_data:
            if cache_status:
                self.log.warn('Trained model cache is not actual because you are using new dataset for training. '
                              'Parameter use_cache value changed to False')
                cache_status = False
        return cache_status

    def fit(self, input_data: InputData, use_cache=True, verbose=False):
        """
        Run training process in all nodes in chain starting with root.

        :param input_data: data used for model training
        :param use_cache: flag defining whether use cache information about previous executions or not, default True
        :param verbose: flag used for status printing to console, default False
        """
        use_cache = self.cache_status_if_new_data(new_input_data=input_data, cache_status=use_cache)

        if not use_cache:
            self._clean_model_cache()

        if input_data.task.task_type == TaskTypesEnum.ts_forecasting:
            # the make_future_prediction is useless for the fit stage
            input_data.task.task_params.make_future_prediction = False
        else:
            if not use_cache or self.fitted_on_data is None:
                self.fitted_on_data = input_data
        train_predicted = self.root_node.fit(input_data=input_data, verbose=verbose)
        return train_predicted

    def predict(self, input_data: InputData):
        """
        Run the predict process in all nodes in chain starting with root.

        :param input_data: data for prediction
        :return: array of predicted target values
        """
        if not self.is_all_cache_actual():
            ex = 'Trained model cache is not actual or empty'
            self.log.error(ex)
            raise Exception(ex)

        result = self.root_node.predict(input_data=input_data)
        return result

    def fine_tune_primary_nodes(self, input_data: InputData, iterations: int = 30,
                                max_lead_time: timedelta = timedelta(minutes=5),
                                verbose=False):
        """
        Optimize hyperparameters in primary nodes models

        :param input_data: data used for tuning
        :param iterations: max number of iterations
        :param max_lead_time: max time available for tuning process
        :param verbose: flag used for status printing to console, default False
        """
        # Select all primary nodes
        # Perform fine-tuning for each model in node
        if verbose:
            self.log.info('Start tuning of primary nodes')

        all_primary_nodes = [node for node in self.nodes if isinstance(node, PrimaryNode)]
        for node in all_primary_nodes:
            node.fine_tune(input_data, max_lead_time=max_lead_time, iterations=iterations)

        if verbose:
            self.log.info('End tuning')

    def fine_tune_all_nodes(self, input_data: InputData, iterations: int = 30,
                            max_lead_time: timedelta = timedelta(minutes=5),
                            verbose=False):
        """
        Optimize hyperparameters in all nodes models

        :param input_data: data used for tuning
        :param iterations: max number of iterations
        :param max_lead_time: max time available for tuning process
        :param verbose: flag used for status printing to console, default False
        """
        if verbose:
            self.log.info('Start tuning of chain')

        node = self.root_node
        node.fine_tune(input_data, max_lead_time=max_lead_time, iterations=iterations)

        if verbose:
            self.log.info('End tuning')

    def add_node(self, new_node: Node):
        """
        Add new node to the Chain

        :param new_node: new Node object
        """
        if new_node not in self.nodes:
            self.nodes.append(new_node)
            if new_node.nodes_from:
                for new_parent_node in new_node.nodes_from:
                    if new_parent_node not in self.nodes:
                        self.add_node(new_parent_node)

    def _actualise_old_node_childs(self, old_node: Node, new_node: Node):
        old_node_offspring = self.node_childs(old_node)
        for old_node_child in old_node_offspring:
            old_node_child.nodes_from[old_node_child.nodes_from.index(old_node)] = new_node

    def replace_node_with_parents(self, old_node: Node, new_node: Node):
        new_node = deepcopy(new_node)
        self._actualise_old_node_childs(old_node, new_node)
        new_nodes = [parent for parent in new_node.ordered_subnodes_hierarchy if not parent in self.nodes]
        old_nodes = [node for node in self.nodes if not node in old_node.ordered_subnodes_hierarchy]
        self.nodes = new_nodes + old_nodes
        self.sort_nodes()

    def update_node(self, old_node: Node, new_node: Node):
        self._actualise_old_node_childs(old_node, new_node)
        new_node.nodes_from = old_node.nodes_from
        self.nodes.remove(old_node)
        self.nodes.append(new_node)
        self.sort_nodes()

    def delete_node(self, node: Node):
        for node_child in self.node_childs(node):
            node_child.nodes_from.remove(node)
        for subtree_node in node.ordered_subnodes_hierarchy:
            self.nodes.remove(subtree_node)

    def _clean_model_cache(self):
        for node in self.nodes:
            node.cache = FittedModelCache(node)

    def is_all_cache_actual(self):
        cache_status = [node.cache.actual_cached_state is not None for node in self.nodes]
        return all(cache_status)

    def node_childs(self, node) -> List[Optional[Node]]:
        return [other_node for other_node in self.nodes if isinstance(other_node, SecondaryNode) and
                node in other_node.nodes_from]

    def _is_node_has_child(self, node) -> bool:
        return any(self.node_childs(node))

    def import_cache(self, fitted_chain: 'Chain'):
        for node in self.nodes:
            if not node.cache.actual_cached_state:
                for fitted_node in fitted_chain.nodes:
                    if fitted_node.descriptive_id == node.descriptive_id:
                        node.cache.import_from_other_cache(fitted_node.cache)
                        break

    # TODO why trees visualisation is incorrect?
    def sort_nodes(self):
        """layer by layer sorting"""
        nodes = self.root_node.ordered_subnodes_hierarchy
        self.nodes = nodes

    def save_chain(self, path: str):
        if not self.template:
            self.template = ChainTemplate(self)
        json_object = self.template.export_to_json(path)
        return json_object

    def load_chain(self, path: str):
        self.nodes = []
        self.log = default_log(__name__)
        self.template = ChainTemplate(self)
        self.template.import_from_json(path)

    def __eq__(self, other) -> bool:
        return self.root_node.descriptive_id == other.root_node.descriptive_id

    @property
    def root_node(self) -> Optional[Node]:
        if len(self.nodes) == 0:
            return None
        root = [node for node in self.nodes
                if not self._is_node_has_child(node)]
        if len(root) > 1:
            raise ValueError(f'{ERROR_PREFIX} More than 1 root_nodes in chain')
        return root[0]

    @property
    def length(self) -> int:
        return len(self.nodes)

    @property
    def depth(self) -> int:
        def _depth_recursive(node):
            if node is None:
                return 0
            if isinstance(node, PrimaryNode):
                return 1
            else:
                return 1 + max([_depth_recursive(next_node) for next_node in node.nodes_from])

        return _depth_recursive(self.root_node)


class SharedChain(Chain):
    def __init__(self, base_chain: Chain, shared_cache: dict, log=None):
        super().__init__(log=log)
        self.nodes = copy(base_chain.nodes)
        for node in self.nodes:
            node.cache = SharedCache(node, global_cached_models=shared_cache)

    def unshare(self) -> Chain:
        chain = Chain()
        chain.nodes = copy(self.nodes)
        for node in chain.nodes:
            node.cache = FittedModelCache(node)
        return chain


def as_nx_graph(chain: Chain):
    graph = nx.DiGraph()
    node_labels = {}
    new_node_idx = {}
    for node in chain.nodes:
        unique_id, label = uuid4(), str(node)
        node_labels[unique_id] = str(node)
        new_node_idx[node] = unique_id
        graph.add_node(unique_id)

    def add_edges(graph, chain, new_node_idx):
        for node in chain.nodes:
            if node.nodes_from is not None:
                for child in node.nodes_from:
                    graph.add_edge(new_node_idx[child], new_node_idx[node])

    add_edges(graph, chain, new_node_idx)
    return graph, node_labels



import warnings
from abc import abstractmethod
from datetime import timedelta
from typing import Optional

from sklearn.cluster import KMeans as SklearnKmeans
from sklearn.discriminant_analysis import (LinearDiscriminantAnalysis,
                                           QuadraticDiscriminantAnalysis)
from sklearn.ensemble import (AdaBoostRegressor,
                              ExtraTreesRegressor,
                              GradientBoostingRegressor,
                              RandomForestClassifier,
                              RandomForestRegressor)
from sklearn.linear_model import (Lasso as SklearnLassoReg,
                                  LinearRegression as SklearnLinReg,
                                  LogisticRegression as SklearnLogReg,
                                  Ridge as SklearnRidgeReg,
                                  SGDRegressor as SklearnSGD)
from sklearn.naive_bayes import BernoulliNB as SklearnBernoulliNB
from sklearn.neighbors import (KNeighborsClassifier as SklearnKNN,
                               KNeighborsRegressor as SklearnKNNReg)
from sklearn.neural_network import MLPClassifier
from sklearn.svm import LinearSVR as SklearnSVR
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from xgboost import XGBClassifier, XGBRegressor

from fedot.core.log import default_log, Log
from fedot.core.models.data import InputData, OutputData
from fedot.core.models.evaluation.custom_models.models import CustomSVC
from fedot.core.models.tuning.hyperparams import params_range_by_model
from fedot.core.models.tuning.tuners import SklearnCustomRandomTuner, SklearnTuner

warnings.filterwarnings("ignore", category=UserWarning)


class EvaluationStrategy:
    """
    Base class to define the evaluation strategy of Model object:
    the certain sklearn or any other model with fit/predict methods.

    :param model_type: str type of the model defined in model repository
    :param dict params: hyperparamters to fit the model with
    :param Log log: Log object to record messages
    """

    def __init__(self, model_type: str, params: Optional[dict] = None,
                 log=None):
        self.params_for_fit = params
        self.model_type = model_type

        if not log:
            self.log: Log = default_log(__name__)
        else:
            self.log: Log = log

    @abstractmethod
    def fit(self, train_data: InputData):
        """
        Main method to train the model with the data provided

        :param InputData train_data: data used for model training
        :return:
        """
        raise NotImplementedError()

    @abstractmethod
    def predict(self, trained_model, predict_data: InputData) -> OutputData:
        """
        Main method to predict the target data.

        :param trained_model: trained model object
        :param InputData predict_data: data to predict
        :return OutputData: passed data with new predicted target
        """
        raise NotImplementedError()

    def fit_tuned(self, train_data: InputData, iterations: int,
                  max_lead_time: timedelta = timedelta(minutes=5)):
        """
        Main method used for hyperparameter searching

        :param train_data: data used for hyperparameter searching
        :param iterations: max number of iterations evaluable for hyperparameter optimization
        :param max_lead_time: max time(seconds) for tuning evaluation
        :return:
        """
        raise NotImplementedError()

    @property
    def implementation_info(self) -> str:
        return 'No description'


class SkLearnEvaluationStrategy(EvaluationStrategy):
    """
    This class defines the certain model implementation for the sklearn models defined in model repository

    :param str model_type: str type of the model defined in model repository
    :param dict params: hyperparameters to fit the model with
    """
    __model_by_types = {
        'xgboost': XGBClassifier,
        'xgbreg': XGBRegressor,
        'adareg': AdaBoostRegressor,
        'gbr': GradientBoostingRegressor,
        'knn': SklearnKNN,
        'knnreg': SklearnKNNReg,
        'dt': DecisionTreeClassifier,
        'dtreg': DecisionTreeRegressor,
        'treg': ExtraTreesRegressor,
        'rf': RandomForestClassifier,
        'rfr': RandomForestRegressor,
        'mlp': MLPClassifier,
        'lda': LinearDiscriminantAnalysis,
        'qda': QuadraticDiscriminantAnalysis,
        'linear': SklearnLinReg,
        'logit': SklearnLogReg,
        'ridge': SklearnRidgeReg,
        'lasso': SklearnLassoReg,
        'kmeans': SklearnKmeans,
        'svc': CustomSVC,
        'svr': SklearnSVR,
        'sgdr': SklearnSGD,
        'bernb': SklearnBernoulliNB
    }

    def __init__(self, model_type: str, params: Optional[dict] = None):
        self._sklearn_model_impl = self._convert_to_sklearn(model_type)
        self._tune_strategy: SklearnTuner = Optional[SklearnTuner]
        super().__init__(model_type, params)

    def fit(self, train_data: InputData):
        """
        This method is used for model training with the data provided

        :param InputData train_data: data used for model training
        :return:
        """
        warnings.filterwarnings("ignore", category=RuntimeWarning)
        if self.params_for_fit:
            sklearn_model = self._sklearn_model_impl(**self.params_for_fit)
        else:
            sklearn_model = self._sklearn_model_impl()

        sklearn_model.fit(train_data.features, train_data.target)
        return sklearn_model

    def predict(self, trained_model, predict_data: InputData) -> OutputData:
        """
        This method used for prediction of the target data.

        :param trained_model: model object
        :param predict_data: data to predict
        :return OutputData: passed data with new predicted target
        """
        raise NotImplementedError()

    def fit_tuned(self, train_data: InputData, iterations: int,
                  max_lead_time: timedelta = timedelta(minutes=5)):
        """
        This method is used for hyperparameter searching

        :param train_data: data used for hyperparameter searching
        :param iterations: max number of iterations evaluable for hyperparameter optimization
        :param max_lead_time: max time(seconds) for tuning evaluation
        :return tuple(object, dict): model with found hyperparameters and dictionary with found hyperparameters
        """
        trained_model = self.fit(train_data=train_data)
        params_range = params_range_by_model.get(self.model_type, None)
        self._tune_strategy = SklearnCustomRandomTuner
        if not params_range:
            self.params_for_fit = None
            return trained_model, trained_model.get_params()

        tuned_params, best_model = self._tune_strategy(trained_model=trained_model,
                                                       tune_data=train_data,
                                                       params_range=params_range,
                                                       cross_val_fold_num=5,
                                                       time_limit=max_lead_time,
                                                       iterations=iterations).tune()

        if best_model or tuned_params:
            self.params_for_fit = tuned_params
            trained_model = best_model

        return trained_model, tuned_params

    def _convert_to_sklearn(self, model_type: str):
        if model_type in self.__model_by_types.keys():
            return self.__model_by_types[model_type]
        else:
            raise ValueError(f'Impossible to obtain SKlearn strategy for {model_type}')

    def _find_model_by_impl(self, impl):
        for model, model_impl in self.__model_by_types.items():
            if model_impl == impl:
                return model

    @property
    def implementation_info(self) -> str:
        return str(self._convert_to_sklearn(self.model_type))


class SkLearnClassificationStrategy(SkLearnEvaluationStrategy):
    def predict(self, trained_model, predict_data: InputData):
        """
        Predict method for classification task

        :param trained_model: model object
        :param predict_data: data used for prediction
        :return: prediction target
        """
        n_classes = len(trained_model.classes_)
        prediction = trained_model.predict_proba(predict_data.features)
        if n_classes < 2:
            raise NotImplementedError()
        elif n_classes == 2:
            prediction = prediction[:, 1]

        return prediction


class SkLearnRegressionStrategy(SkLearnEvaluationStrategy):
    def predict(self, trained_model, predict_data: InputData):
        """
        Predict method for regression task

        :param trained_model: model object
        :param predict_data: data used for prediction
        :return:
        """
        prediction = trained_model.predict(predict_data.features)
        return prediction


class SkLearnClusteringStrategy(SkLearnEvaluationStrategy):
    def fit(self, train_data: InputData):
        """
        Fit method for clustering task

        :param train_data: data used for model training
        :return:
        """
        sklearn_model = self._sklearn_model_impl(n_clusters=2)
        sklearn_model = sklearn_model.fit(train_data.features)
        return sklearn_model

    def predict(self, trained_model, predict_data: InputData) -> OutputData:
        """
        Predict method for clustering task
        :param trained_model: model object
        :param predict_data: data used for prediction
        :return:
        """
        prediction = trained_model.predict(predict_data.features)
        return prediction

    def fit_tuned(self, train_data: InputData, iterations: int = 30,
                  max_lead_time: timedelta = timedelta(minutes=5)):
        """
        This method is used for hyperparameter searching

        :param train_data: data used for hyperparameter searching
        :param iterations: max number of iterations evaluable for hyperparameter optimization
        :param max_lead_time: max time(seconds) for tuning evaluation
        :return:
        """
        raise NotImplementedError()
